import pandas as pd
import traceback
from ...wrapper.mysql import (RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector, 
    ViewDatabaseConnector)
from ...view.view_models import FundDailyCollection
from ...view.raw_models import RqFundIndicator
from ...view.basic_models import FundNav, FundInfo, FundRatingLatest, FundSize
from ...view.derived_models import FundAlpha, FundScore, FundIndicator
from ....constant import ExchangeStatus

class FundDailyCollectionProcessor(object):
    def get_fund_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundInfo.fund_id,
                    FundInfo.order_book_id,
                    FundInfo.wind_class_1,
                    FundInfo.start_date,
                    FundInfo.manager_id,
                    FundInfo.company_id,
                    FundInfo.benchmark,
                    FundInfo.desc_name,
                    FundInfo.track_index,
                )
                funds = pd.read_sql(query.statement, query.session.bind)
                return funds.set_index('fund_id')
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_nav_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundNav).order_by(FundNav.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                query = mn_session.query(
                    FundNav.fund_id,
                    FundNav.unit_net_value,
                    FundNav.acc_net_value,
                    FundNav.adjusted_net_value,
                    FundNav.datetime,
                    FundNav.redeem_status,
                    FundNav.subscribe_status,
                ).filter(FundNav.datetime==latest_time)

                nav = pd.read_sql(query.statement, query.session.bind)
                return nav.set_index('fund_id'), latest_time
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    # TODO: use data in basic
    def get_indicator_df(self):
        with RawDatabaseConnector().managed_session() as quant_session:
            try:
                latest_time = quant_session.query(RqFundIndicator).order_by(RqFundIndicator.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                query = quant_session.query(
                    RqFundIndicator.order_book_id,
                    RqFundIndicator.annualized_returns,
                    RqFundIndicator.annualized_risk,
                    RqFundIndicator.average_size,
                    RqFundIndicator.information_ratio,
                    RqFundIndicator.last_month_return,
                    RqFundIndicator.last_six_month_return,
                    RqFundIndicator.last_three_month_return,
                    RqFundIndicator.last_twelve_month_return,
                    RqFundIndicator.last_week_return,
                    RqFundIndicator.year_to_date_return,
                    RqFundIndicator.to_date_return,
                    RqFundIndicator.sharp_ratio,
                    RqFundIndicator.max_drop_down,
                ).filter(RqFundIndicator.datetime==latest_time)
                indicator = pd.read_sql(query.statement, query.session.bind)
                return indicator

            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_institution_rating_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundRatingLatest).order_by(FundRatingLatest.update_time.desc()).limit(1).one_or_none()
                latest_time = latest_time.update_time
                print(latest_time)
                query = mn_session.query(
                    FundRatingLatest.fund_id,
                    FundRatingLatest.zs,
                    FundRatingLatest.sh3,
                    FundRatingLatest.sh5,
                    FundRatingLatest.jajx,
                ).filter(FundRatingLatest.update_time==latest_time)
                rating = pd.read_sql(query.statement, query.session.bind)
                return rating

            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_size_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundSize.fund_id,
                    FundSize.latest_size
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.drop_duplicates(subset=['fund_id'], keep='last')
                df = df.set_index('fund_id')
                return df
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_alpha_df(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundAlpha).order_by(FundAlpha.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    FundAlpha.fund_id,
                    FundAlpha.track_err,
                    FundAlpha.this_y_alpha,
                    FundAlpha.cumulative_alpha,
                    FundAlpha.w1_alpha,
                    FundAlpha.m1_alpha,
                    FundAlpha.m3_alpha,
                    FundAlpha.m6_alpha,
                    FundAlpha.y1_alpha,
                    FundAlpha.y3_alpha,
                    FundAlpha.y5_alpha,
                    FundAlpha.y10_alpha,
                ).filter(FundAlpha.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.drop_duplicates(subset=['fund_id'], keep='last')
                return df

            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_score_df(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundScore.datetime).order_by(FundScore.datetime.desc()).limit(1).one_or_none()[0]
                print(latest_time)
                query = mn_session.query(
                    FundScore.fund_id,
                    FundScore.tag_name,
                    FundScore.score,
                ).filter(FundScore.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.drop_duplicates(subset=['fund_id'], keep='last')
                return df

            except Exception as e:
                print('Failed get_score_df <err_msg> {}'.format(e))

    def get_tag_fund_indicator_df(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundIndicator).order_by(FundIndicator.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    FundIndicator.fund_id,
                    FundIndicator.alpha,
                    FundIndicator.beta,
                    FundIndicator.track_err,
                    FundIndicator.fee_rate,
                ).filter(FundIndicator.datetime==latest_time)
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={'track_err': 'tag_track_err'})
                df = df.drop_duplicates(subset=['fund_id'], keep='last')
                return df

            except Exception as e:
                print('Failed get_fund_indicator_df <err_msg> {}'.format(e))

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def get_exchange_status(self, x):
        if x['redeem_status'] == ExchangeStatus.LIMITED:
            x['redeem_status'] = ExchangeStatus.OPEN
        if x['subscribe_status'] == ExchangeStatus.LIMITED:
            x['subscribe_status'] = ExchangeStatus.OPEN
        if x['redeem_status'] == ExchangeStatus.CLOSE and x['subscribe_status'] == ExchangeStatus.CLOSE:
            return '暂停交易'
        if x['redeem_status'] == ExchangeStatus.SUSPENDED:
            return '暂停赎回'
        if x['subscribe_status'] == ExchangeStatus.SUSPENDED:
            return '暂停申购'
        if x['redeem_status'] == ExchangeStatus.OPEN and x['subscribe_status'] == ExchangeStatus.OPEN:
            return '正常开放'
        return '暂无状态'

    def get_found_to_now(self, x):
        if pd.isna(x):
            return None
        return round((pd.datetime.now().date() - x).days / 365, 2)

    def modify_fund_manager(self, x):
        try:
            return x.split('\r\n')[-1].split('(')[0]
        except:
            return x

    def compute_daily_collection(self):
        try:
            funds = self.get_fund_df()
            funds = funds.reset_index()
            funds = funds.drop_duplicates(subset=['fund_id'], keep='last')
            funds = funds.set_index('fund_id')
            print('info', funds)
            print('-' * 100)

            nav, latest_time = self.get_nav_df()
            print(latest_time)
            nav = nav.reset_index()
            nav = nav.drop_duplicates(subset=['fund_id'], keep='last')
            nav = nav.set_index('fund_id')
            print('nav', nav)
            print('-' * 100)

            alpha = self.get_alpha_df()
            alpha = alpha.set_index('fund_id')
            print(alpha)
            print('-' * 100)
            size = self.get_size_df()
            print(size)
            print('-' * 100)
            score = self.get_score_df().set_index('fund_id')
            print(score)
            print('-' * 100)
            tag_indicator = self.get_tag_fund_indicator_df().set_index('fund_id')
            print(tag_indicator)
            print('-' * 100)

            print('处理机构评级')
            rating = self.get_institution_rating_df()
            rating = rating.drop_duplicates(subset=['fund_id'], keep='last')
            rating = rating.set_index('fund_id')
            print(rating)
            print('-' * 100)

            print('处理净值')
            df = pd.concat([funds, nav, alpha, size, rating, score, tag_indicator], axis=1, sort=False)
            df.index.name = 'fund_id'
            print('_'*100)
            print(df)

            print('处理行情')
            indicator = self.get_indicator_df()
            indicator = indicator.dropna(subset=['order_book_id'])
            indicator['fund_id'] = indicator['order_book_id'] + '!0'
            indicator = indicator.drop_duplicates(subset=['order_book_id'], keep='last')
            indicator = indicator.drop(['order_book_id'], axis=1)
            indicator = indicator.set_index('fund_id')
            print(indicator)
            print('-' * 100)

            df = pd.concat([df, indicator.reindex(index=df.index)], axis=1, sort=False)
            df.index.name = 'fund_id'
            df = df.reset_index()
            df = df.dropna(subset=['fund_id'])
            df['exchange_status'] = df.apply(self.get_exchange_status, axis=1)
            df['found_to_now'] = df['start_date'].apply(self.get_found_to_now)
            df['manager_id'] = df['manager_id'].apply(self.modify_fund_manager)
            df = df.rename(columns={
                'wind_class_1': 'wind_class_I',
                'start_date': 'found_date',
                'manager_id': 'fund_manager',
                'company_id': 'company_name',
                'desc_name': 'symbol',
            })
            df = df.drop(['redeem_status', 'subscribe_status'], axis=1)
            df = df.sort_values(['fund_id'])
            df = df.drop_duplicates(subset=['order_book_id'], keep='first')
            df = df.dropna(subset=['order_book_id'])

            # df.to_csv('./hhh.csv')
            self.append_data(FundDailyCollection.__tablename__, df)

            print(df)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def process(self):
        failed_tasks = []
        if not self.compute_daily_collection():
            failed_tasks.append('compute_daily_collection')
        return failed_tasks


if __name__ == '__main__':
    FundDailyCollectionProcessor().compute_daily_collection()





from .rq_raw_data_downloader import RqRawDataDownloader
from .em_raw_data_downloader import EmRawDataDownloader
from .web_raw_data_downloader import WebRawDataDownloader

class RawDataDownloader(object):
    def __init__(self, rq_license):
        self.rq_downloader = RqRawDataDownloader(rq_license)
        self.web_downloader = WebRawDataDownloader()
        self.em_downloader = EmRawDataDownloader()
        self.updated_count = {}

    def download(self, start_date, end_date):
        failed_tasks = []

        failed_tasks.extend(self.em_downloader.download_all(start_date, end_date))
        # If 'em_tradedates' in failed_tasks, there is no trading day between start_date and end_date
        # Stop and return
        if 'em_tradedates' in failed_tasks:
            return failed_tasks

        failed_tasks.extend(self.rq_downloader.download_all(start_date, end_date))
        failed_tasks.extend(self.web_downloader.download_all(start_date, end_date))
        
        for key, value in self.rq_downloader.updated_count.items():
            if key not in self.updated_count:
                self.updated_count[key] = value
            else:
                self.updated_count[key] += value

        for key, value in self.web_downloader.updated_count.items():
            if key not in self.updated_count:
                self.updated_count[key] = value
            else:
                self.updated_count[key] += value

        for key, value in self.em_downloader.updated_count.items():
            if key not in self.updated_count:
                self.updated_count[key] = value
            else:
                self.updated_count[key] += value

        return failed_tasks



from .basic_data_part1 import BasicDataPart1
from .basic_data_part2_fund_ret import BasicFundRet

class BasicDataProcessor(object):
    def __init__(self):
        self.basic_data_part1 = BasicDataPart1()
        self.basic_fund_ret = BasicFundRet()
        self.updated_count = {}

    def process_all(self, start_date, end_date):
        failed_tasks = []

        failed_tasks.extend(self.basic_data_part1.process_all(start_date, end_date))
        failed_tasks.extend(self.basic_fund_ret.process_all(end_date))

        for key, value in self.basic_data_part1.updated_count.items():
            if key not in self.updated_count:
                self.updated_count[key] = value
            else:
                self.updated_count[key] += value

        for key, value in self.basic_fund_ret.updated_count.items():
            if key not in self.updated_count:
                self.updated_count[key] = value
            else:
                self.updated_count[key] += value

        return failed_tasks


from .fund_score_processor import FundScoreProcessor
from .fund_indicator_processor import FundIndicatorProcessor
from .derived_index_val import IndexValProcess

class DerivedDataProcessor(object):
    def __init__(self):
        self.fund_indicator_processor = FundIndicatorProcessor()
        self.fund_score_processor = FundScoreProcessor()
        self.index_val_processor = IndexValProcess()
        self.updated_count = {}

    def process_all(self, start_date, end_date):
        failed_tasks = []

        failed_tasks.extend(self.fund_indicator_processor.process(start_date, end_date))
        failed_tasks.extend(self.fund_score_processor.process(start_date, end_date))
        failed_tasks.extend(self.index_val_processor.process(start_date, end_date))
        
        for key, value in self.fund_indicator_processor.updated_count.items():
            if key not in self.updated_count:
                self.updated_count[key] = value
            else:
                self.updated_count[key] += value

        for key, value in self.fund_score_processor.updated_count.items():
            if key not in self.updated_count:
                self.updated_count[key] = value
            else:
                self.updated_count[key] += value

        for key, value in self.index_val_processor.updated_count.items():
            if key not in self.updated_count:
                self.updated_count[key] = value
            else:
                self.updated_count[key] += value

        return failed_tasks

if __name__ == '__main__':
    ddp = DerivedDataProcessor()
    # start_date = '20200430'
    start_date = '20200506'
    end_date = '20200506'
    ddp.fund_score_processor.process(start_date, end_date)



import pandas as pd
import traceback
import datetime
from ...wrapper.mysql import BasicDatabaseConnector
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.basic_models import *

class BasicDataPart1(object):

    def __init__(self):
        self._index_info_df = BasicDataApi().get_index_info()
        self._fund_info_df = BasicDataApi().get_fund_info()
        # Get the mapping between order_book_id and fund_id (which is defined in this system)
        self._fund_id_mapping = BasicDataApi().get_fund_id_mapping()
        self.updated_count = {}

    def _upload_basic(self, df, table_name, to_truncate=False):
        print(table_name)
        # print(df)
        if to_truncate:
            with BasicDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
        df.to_sql(table_name, BasicDatabaseConnector().get_engine(), index=False, if_exists='append')

        if table_name not in self.updated_count:
            self.updated_count[table_name] = 0
        self.updated_count[table_name] += df.shape[0]

    def _get_fund_id_from_order_book_id(self, order_book_id, date):
        if order_book_id in self._fund_id_mapping:
            for fund_info in self._fund_id_mapping[order_book_id]:
                if (date >= fund_info['start_date'] and
                        (fund_info['end_date'] == '00000000' or date <= fund_info['end_date'])):
                    return fund_info['fund_id']
        return None

    def _status_mapping(self, x):
        if x == 'Open':
            return 0
        elif x == 'Suspended':
            return 1
        elif x == 'Limited':
            return 2
        elif x == 'Close':
            return 3
        else:
            return None

    def _find_tag(self, symbol, wind_class_II):
        if '沪深300' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return 'A股大盘'
        elif '中证500' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return 'A股中盘'
        elif '标普500' in symbol:
            return '美股大盘'
        elif '创业板' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return '创业板'
        elif '德国' in symbol:
            return '德国大盘'
        elif '日本' in symbol or '日经' in symbol:
            return '日本大盘'
        elif (('国债' in symbol or '利率' in symbol or '金融债' in symbol)
              and wind_class_II in ['短期纯债型基金', '中长期纯债型基金', '被动指数型债券基金']):
            return '利率债'
        elif (('信用' in symbol or '企债' in symbol or '企业债' in symbol)
              and wind_class_II in ['短期纯债型基金', '中长期纯债型基金', '被动指数型债券基金']):
            return '信用债'
        elif '黄金' in symbol:
            return '黄金'
        elif '原油' in symbol or '石油' in symbol or '油气' in symbol:
            return '原油'
        elif ('地产' in symbol or '金融' in symbol) and ('美国'not in symbol):
            return '房地产'
        else:
            return 'null'

    def fund_info(self):
        # Update manually
        # Not verified
        track_index_df = pd.read_csv('./data/fund_track_index.csv', index_col=0)
        wind_fund_info = RawDataApi().get_wind_fund_info()
        fund_fee = RawDataApi().get_fund_fee()
        wind_fund_info['order_book_id'] = [_.split('!')[0].split('.')[0] for _ in wind_fund_info['wind_id'] ]
        res = []
        for i in wind_fund_info['wind_id']:
            if not '!' in i:
                res.append(0)
            else:
                res_i = int(i.split('!')[1].split('.')[0])
                res.append(res_i)
        wind_fund_info['transition'] = res
        wind_fund_info['fund_id'] = [o + '!' + str(t) for o, t in zip(wind_fund_info['order_book_id'], wind_fund_info['transition'])]
        wind_fund_info = wind_fund_info.set_index('fund_id')
        fund_fee = fund_fee.drop(['id','desc_name'], axis = 1).set_index('fund_id')
        wind_fund_info = wind_fund_info.join(fund_fee)
        wind_fund_info['fund_id'] = wind_fund_info.index
        wind_fund_info['asset_type'] = [self._find_tag(symbol, wind_class_II) for symbol, wind_class_II in zip(wind_fund_info['desc_name'],wind_fund_info['wind_class_2'])]
        wind_fund_info = wind_fund_info[[i.split('.')[1] == 'OF' for i in wind_fund_info['wind_id']]]
        wind_fund_info = wind_fund_info.set_index('wind_id')
        dic = {k:v for k,v in zip(self._index_info_df['desc_name'], self._index_info_df['index_id'])}
        wind_fund_info = wind_fund_info.join(track_index_df[['track_index']])
        wind_fund_info['wind_id'] = wind_fund_info.index
        wind_fund_info['track_index'] = wind_fund_info['track_index'].map(lambda x: dic.get(x,'none'))
        self._upload_basic(wind_fund_info, FundInfo.__table__.name)

    def index_info(self):
        # Update manually
        # Not verified
        df = pd.read_csv('./data/index_info.csv')
        self._upload_basic(df, IndexInfo.__table__.name)

    def fund_nav_from_rq(self, start_date, end_date):
        df = RawDataApi().get_rq_fund_nav(start_date, end_date)
        df['fund_id'] = df.apply(
            lambda x: self._get_fund_id_from_order_book_id(x['order_book_id'], x['datetime']), axis=1)
        df['subscribe_status'] = df['subscribe_status'].map(self._status_mapping)
        df['redeem_status'] = df['redeem_status'].map(self._status_mapping)
        df = df[df['fund_id'].notna()]
        self._upload_basic(df, FundNav.__table__.name)

    def fund_nav(self, start_date, end_date):
        try:
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d') - datetime.timedelta(days=3)
            start_date = start_date.strftime('%Y%m%d')
            df = RawDataApi().get_em_fund_nav(start_date, end_date)
            df['fund_id'] = df.apply(
                lambda x: self._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
            df = df.drop(['CODES'], axis=1)
            df = df[df['fund_id'].notna()]
            df = df.rename(columns={
                    'DATES': 'datetime',
                    'ORIGINALUNIT': 'unit_net_value',
                    'ORIGINALNAVACCUM': 'acc_net_value',
                    'ADJUSTEDNAV': 'adjusted_net_value',
                    'UNITYIELD10K': 'daily_profit',
                    'YIELDOF7DAYS': 'weekly_yield'
                })
            df['change_rate'] = df.adjusted_net_value / df.adjusted_net_value.shift(1) - 1
            df['weekly_yield'] = df['weekly_yield'].map(lambda x: x / 100.0)
            end_date = datetime.datetime.strptime(end_date, '%Y%m%d').date()
            df = df[df.datetime == end_date]
            self._upload_basic(df, FundNav.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def index_price(self, start_date, end_date):
        try:
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d') - datetime.timedelta(days=3)
            start_date = start_date.strftime('%Y%m%d')
            cm_index_price = RawDataApi().get_raw_cm_index_price_df(start_date, end_date)
            cxindex_index_price = RawDataApi().get_cxindex_index_price_df(start_date, end_date)
            yahoo_index_price = RawDataApi().get_yahoo_index_price_df(start_date, end_date)
            rq_index_price = RawDataApi().get_rq_index_price_df(start_date, end_date)

            df_list = []
            index_list = ['sp500', 'dax30', 'n225']
            cm_index_list = ['sp500rmb', 'dax30rmb', 'n225rmb']
            for i, c in zip(index_list, cm_index_list):
                cm_index = yahoo_index_price.copy()
                df = cm_index[cm_index['index_id'] == i]
                df = cm_index_price.set_index('datetime').join(df.drop(['_update_time'], axis=1).set_index('datetime'))
                df = df.fillna(method='ffill')
                df['close'] = df['close'] * df['usd_central_parity_rate']
                df = df[['close']]
                df['datetime'] = df.index
                df['open'] = float('Nan')
                df['high'] = float('Nan')
                df['low'] = float('Nan')
                df['volume'] = float('Nan')
                df['total_turnover'] = float('Nan')
                df['index_id'] = c
                df['ret'] = df.close / df.close.shift(1) -1
                df_list.append(df.copy())

            index_dic = {k: v for k, v in zip(
                self._index_info_df['order_book_id'], self._index_info_df['index_id']) if k != 'not_available'}
            rq_index_price['index_id'] = rq_index_price['order_book_id'].map(
                lambda x: index_dic[x])
            res = []
            for index_i in list(set(rq_index_price['order_book_id'].tolist())):
                # update gold index data from em not rq
                if index_dic[index_i] == 'gold':
                    continue 
                dftmp = rq_index_price[rq_index_price['order_book_id'] == index_i].copy()
                dftmp['ret'] = dftmp.close / dftmp.close.shift(1) - 1
                res.append(dftmp)
            rq_index_price = pd.concat(res)
            df_list.append(yahoo_index_price)
            df_list.append(cxindex_index_price)
            df_list.append(rq_index_price)

            df = pd.concat(df_list).drop(['order_book_id'], axis=1)
            df['datetime'] = pd.to_datetime(df['datetime'])
            end_date = datetime.datetime.strptime(end_date, '%Y%m%d').date()
            df = df[df.datetime==end_date]
            self._upload_basic(df, IndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def fund_rating_latest(self):
        try:
            fund_rating_df = RawDataApi().get_fund_rating()
            fund_info = self._fund_info_df[['fund_id', 'order_book_id', 'start_date', 'end_date']]
            score_name = ['zs', 'sh3', 'sh5', 'jajx']
            score_dic = {}
            for s in score_name:
                score_df = fund_rating_df[['order_book_id', 'datetime', s]]
                datelist = sorted(list(set(score_df.dropna().datetime.tolist())))
                score_dic[s] = datelist[-1]
            res = []
            for s, d in score_dic.items():
                try:
                    df = fund_rating_df[['order_book_id', 'datetime', s]]
                    df = df[df['datetime'] == d]
                    con1 = fund_info['start_date'] <= d
                    con2 = fund_info['end_date'] >= d
                    fund_info_i = fund_info[con1 & con2]
                    dic = {row['order_book_id']: row['fund_id']
                        for index, row in fund_info_i.iterrows()}
                    df['fund_id'] = df['order_book_id'].map(lambda x: dic[x])
                    df = df[['fund_id', s]].copy().set_index('fund_id')
                    res.append(df)
                except Exception as e:
                    print(e)
            df = pd.concat(res, axis=1, sort=False)
            df['fund_id'] = df.index
            df['update_time'] = datetime.date.today()
            self._upload_basic(df, FundRatingLatest.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def rq_stock_price(self, start_date, end_date):
        try:
            stock_price_df = RawDataApi().get_rq_stock_price(start_date, end_date)
            stock_post_price_df = RawDataApi().get_rq_stock_post_price(start_date, end_date)

            stock_post_price_df['adj_close'] = stock_post_price_df['close']
            stock_post_price_df = stock_post_price_df.filter(items=['adj_close', 'datetime', 'order_book_id'], 
                axis='columns')

            stock_price_merge_df = stock_price_df.merge(stock_post_price_df, how='left', on=['datetime', 'order_book_id'])
            stock_price_merge_df['post_adj_factor'] = stock_price_merge_df['adj_close'] / stock_price_merge_df['close']
            stock_price_merge_df = stock_price_merge_df.rename(columns={'order_book_id': 'stock_id'})
            self._upload_basic(stock_price_merge_df, StockPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def rq_fund_ret(self, start_date, end_date):
        try:
            column_rename_dict = {
                'sharp_ratio': 'sharpe_ratio',
                'last_week_return': 'w1_ret',
                'last_month_return': 'm1_ret',
                'last_three_month_return': 'm3_ret',
                'last_six_month_return': 'm6_ret',
                'last_twelve_month_return': 'y1_ret',
                'max_drop_down': 'mdd',
                'annualized_returns': 'annual_ret',
                'average_size': 'avg_size',
                'information_ratio': 'info_ratio',
                'to_date_return':'to_date_ret'
            }
            column_drop_list = ['order_book_id', 'year_to_date_return', 'annualized_risk']

            df = RawDataApi().get_rq_fund_indicator(start_date, end_date)
            df['fund_id'] = df.apply(
                lambda x: self._get_fund_id_from_order_book_id(x['order_book_id'], x['datetime']), axis=1)
            df = df.rename(columns=column_rename_dict).drop(columns=column_drop_list)
            df = df[df['fund_id'].notnull()]

            self._upload_basic(df, FundRet.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def stock_info(self):
        try:
            df = RawDataApi().get_stock_info()
            self._upload_basic(df, StockInfo.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def fund_size(self, start_date, end_date):
        try:
            # Since fund_size in basic db is latest snapshot, we only use end_date as param
            df = RawDataApi().get_em_fund_scale(end_date, end_date)
            df['fund_id'] = df.apply(
                lambda x: self._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
            df = df.drop(['CODES', 'DATES'], axis = 1).rename(columns={'FUNDSCALE': 'latest_size'})
            df = df[df['fund_id'].notnull()]
            self._upload_basic(df, FundSize.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def trading_day_list(self, start_date, end_date):
        try:
            df = RawDataApi().get_em_tradedates(start_date, end_date)
            df = df.rename(columns={'TRADEDATES': 'datetime'})
            self._upload_basic(df, TradingDayList.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False
    
    def em_index_price(self, start_date, end_date):
        try:
            start_date = datetime.datetime.strptime(start_date, '%Y%m%d') - datetime.timedelta(days=5)
            del_date = start_date + datetime.timedelta(days=1)
            start_date = start_date.strftime('%Y%m%d')
            index_info = BasicDataApi().get_index_info()
            em_id_list = index_info['em_id'].dropna().tolist()
            raw_index_price = RawDataApi().get_em_index_price(start_date, end_date, em_id_list)
            res = [] 
            index_id_list = []
            for em_id in em_id_list:
                df = raw_index_price[raw_index_price['em_id'] == em_id].copy()
                index_id = index_info[index_info['em_id'] == em_id].index_id.values[0]
                df.drop(['em_id'], axis = 1, inplace=True)
                df['index_id'] = index_id
                df['high'] = float('nan')
                df['low'] = float('nan')
                df['open'] = float('nan')
                df['total_turnover'] = float('nan')
                df['volume'] = float('nan')
                df['ret'] = df.close / df.close.shift(1) - 1
                res.append(df)
                if not df.empty:
                    index_id_list.append(index_id)
            BasicDataApi().delete_index_price(index_id_list, del_date, end_date)
            df = pd.concat(res, axis=0, sort=False)
            df = df[df.datetime >= del_date.date()]
            self._upload_basic(df, IndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def process_all(self, start_date, end_date):
        failed_tasks = []

        # Depend on RQ
        # if not self.rq_fund_ret(start_date, end_date):
        #     failed_tasks.append('rq_fund_ret')

        # # Depend on RQ
        # if not self.stock_info():
        #     failed_tasks.append('stock_info')

        # Depend on RQ
        # if not self.rq_stock_price(start_date, end_date):
        #     failed_tasks.append('rq_stock_price')

        # TODO: Depend on RQ, change to EM
        if not self.index_price(start_date, end_date):
            failed_tasks.append('index_price')

        # TODO: Depend on RQ, change to EM
        if not self.fund_rating_latest():
            failed_tasks.append('fund_rating_latest')
            
        if not self.fund_size(start_date, end_date):
            failed_tasks.append('fund_size')

        if not self.fund_nav(start_date, end_date):
            failed_tasks.append('fund_nav')

        if not self.em_index_price(start_date, end_date):
            failed_tasks.append('em_index_price')

        if not self.trading_day_list(start_date, end_date):
            failed_tasks.append('trading_day_list')

        return failed_tasks

if __name__ == '__main__':
    bdp = BasicDataPart1()
    # start_date = '20200428'
    # end_date = '20200428'
    # bdp.index_price(start_date, end_date)
    # bdp.em_index_price(start_date, end_date)
    # bdp.rq_stock_price(start_date, end_date)
    # bdp.rq_fund_ret(start_date, end_date)
    # bdp.fund_rating_latest()
    bdp.fund_size('20200513', '20200513')



import datetime
import pandas as pd
import numpy as np
import traceback
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...wrapper.mysql import DerivedDatabaseConnector
from ...view.derived_models import IndexValuationDevelop

class IndexValProcess:

    YEAR = 242
    WINDOW      = YEAR * 5
    MIN_PERIOD  = YEAR * 3
    SCORE_METHOD = {
        'PB百分位': 'pb_pct',
        'PE百分位': 'pe_pct',
        'PS百分位': 'ps_pct',
    }

    def __init__ (self):
        self.raw_api = RawDataApi()
        self.basic_api = BasicDataApi()
        self.updated_count = {}

    def init(self, start_time, end_time):
        self.start_time = datetime.datetime.strptime(start_time, '%Y%m%d').date()
        self.end_time = datetime.datetime.strptime(end_time, '%Y%m%d').date()
        self.index_info = self.basic_api.get_index_info()
        em_id_to_index_id = self.index_info[['index_id','em_id']].set_index('em_id').to_dict()['index_id']
        em_index_list = self.index_info[self.index_info['tag_method'] != 'none'].em_id
        self.index_val = self.raw_api.get_em_index_val(self.start_time, self.end_time, em_index_list)   
        self.index_val['index_id'] = self.index_val['em_id'].map(lambda x : em_id_to_index_id[x])
        self.index_val = self.index_val.drop(['em_id','_update_time'], axis = 1)
        self.index_val = self.index_val.set_index(['index_id','datetime'])
        self.index_id_score_method_dict = self.index_info[self.index_info['tag_method'] != 'none'][['index_id','tag_method']].set_index('index_id').to_dict()['tag_method']

    def upload_derived(self, df, table_name):
        print(table_name)
        df.to_sql(table_name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')

        if table_name not in self.updated_count:
            self.updated_count[table_name] = 0
        self.updated_count[table_name] += df.shape[0]

    def calculate(self):
        pctrank = lambda x: pd.Series(x).rank(pct=True).iloc[-1]
        self.index_val = self.index_val.rename(columns = {'dividend_yield' : 'dy'})
        index_list = self.index_val.index.levels[0].tolist()
        self.result = []
        for index_id in index_list:
            df_i = self.index_val.loc[index_id].copy()
            df_i.loc[:,'pe_pct'] = df_i['pe_ttm'].rolling(window= self.WINDOW, min_periods=self.MIN_PERIOD ).apply(pctrank, raw=True)
            df_i.loc[:,'pb_pct'] = df_i['pb_mrq'].rolling(window= self.WINDOW, min_periods=self.MIN_PERIOD ).apply(pctrank, raw=True)
            df_i.loc[:,'ps_pct'] = df_i['ps_ttm'].rolling(window= self.WINDOW, min_periods=self.MIN_PERIOD ).apply(pctrank, raw=True)
            df_i.loc[:,'roe_pct'] = df_i['roe'].rolling(window= self.WINDOW, min_periods=self.MIN_PERIOD ).apply(pctrank, raw=True)
            df_i['peg_ttm'] = df_i['pe_ttm'] / (df_i['dy'] / df_i['dy'].shift(self.YEAR) -1 ) 
            df_i.loc[:,'index_id'] = index_id
            self.result.append(df_i.copy())
        self.result = pd.concat(self.result).reset_index()
        score_list = []
        for idx, r in self.result.iterrows():
            tag_method = self.index_id_score_method_dict[r.index_id]
            score_data = self.SCORE_METHOD[tag_method]
            score_list.append(r[score_data])
        
        self.result['datetime'] = pd.to_datetime(self.result['datetime'])
        self.result['val_score'] = score_list    
        self.result = self.result.replace(np.inf, float('nan')).replace(-np.inf, float('nan'))
        
    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            start_date_dt = datetime.datetime.strptime(start_date, '%Y%m%d').date()
            end_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()
            start_date = start_date_dt - datetime.timedelta(days = 2000) #5年历史保险起见，多取几天 5*365=1825
            start_date = datetime.datetime.strftime(start_date, '%Y%m%d')
            self.init(start_date, end_date)
            self.calculate()
            self.result = self.result[self.result['datetime'] >= pd.to_datetime(start_date_dt)]
            self.result = self.result[self.result['datetime'] <= pd.to_datetime(end_date_dt)]
            self.upload_derived(self.result, IndexValuationDevelop.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_stack()
            failed_tasks.append('index_val')

        return failed_tasks



import pandas as pd
import numpy as np
import platform
import matplotlib as mpl
import pylab as pl
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.dates as mdates
from matplotlib.patches import Rectangle
from ....data.manager.score import FundScoreManager
CURRENT_PLATFORM = platform.system()
if CURRENT_PLATFORM == 'Darwin':
    mpl.rcParams['font.family'] = ['Heiti TC']
else:
    mpl.rcParams['font.family'] = ['STKaiti']

class FundPainter(object):

    @staticmethod
    def plot_fund_weights(  fund_weights_history:dict, 
                            fund_cash_history:dict, 
                            fund_marktet_price_history:dict):
        res = []
        for k,v in fund_weights_history.items():
            v = v
            v['date'] = k
            v['cash'] = fund_cash_history[k]  / fund_marktet_price_history[k]
            res.append(v)
        weights_df = pd.DataFrame(res).set_index('date')
        weights_df = weights_df.div(weights_df.sum(axis=1), axis=0)
        weights_df.fillna(0)[1:].plot.area(figsize=(18,9),legend=False,fontsize = 13)
        s = pl.title('fund weights history', fontsize=20)

    @staticmethod
    def plot_index_fund_fee( index_fee:pd.DataFrame):
        mpl.rcParams['font.size'] = 15
        p = index_fee.plot.pie(y='amount',figsize = (10,10))
        l = plt.legend(fontsize=13 ,loc = 'lower left')
        t = plt.title('index fee amount', fontsize=20)
        y = plt.ylabel(ylabel='fee amount', fontsize=14)

    @staticmethod
    def plot_fund_mdd_periods(  fund_mv:pd.DataFrame, 
                                fund_weights_history:dict, 
                                fund_nav:pd.DataFrame, 
                                fund_info:pd.DataFrame):
        df = fund_mv
        mdd_part1 = (df.loc[:, 'mv'] / df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        mdd_date1 = df.loc[:mdd_part1.idxmin(),'mv'].idxmax()
        mdd_date2 = mdd_part1.idxmin()
        date_list = np.array(list(fund_weights_history.keys()))
        date_list = date_list[(date_list >= mdd_date1) & (date_list <= mdd_date2)]
        fund_list = []
        for d in date_list:
            fund_list.extend([k for k,v in fund_weights_history[d].items() if isinstance(v, float) and (round(v,3) > 0)])
        fund_list = list(set(fund_list)) 
        if 'cash' in fund_list:
            fund_list.remove('cash')
        df = fund_nav[fund_list].loc[mdd_date1:mdd_date2,:] 
        if df.empty:
            return 
        fig, ax = plt.subplots(figsize= [18,9])
        df = df/df.iloc[0]
        table_df = fund_info.set_index('fund_id').loc[fund_list,:].reset_index()[['fund_id','desc_name','index_id']].sort_values(['index_id'])
        df_mdd = pd.DataFrame(((1-df.iloc[-1])*100).round(2))
        df_mdd.columns = ['max draw down %']
        table_df = table_df.set_index('fund_id').join(df_mdd).reset_index()
        fund_list = table_df.fund_id.tolist()
        
        for col in df.columns:
            plt.plot(df.index, df[[col]], linewidth=1.0, label = col)
        plt.legend(fontsize=13 ,loc = 'lower left')
        plt.title('fund bt during mdd period  nav of all funds', fontsize=20)
        ax.xaxis.set_ticks_position('top')
        t = plt.table(cellText=table_df.values.tolist(),
          colLabels=table_df.columns,
          colWidths= [0.25,0.45,0.15,0.15],  
          loc='bottom',
          )
        t.auto_set_font_size(False)
        t.auto_set_font_size(False)
        t.set_fontsize(15)
        t.auto_set_column_width('fund_id')
        t.AXESPAD = 0.1
        t.scale(1, 2)
        plt.show()

    @staticmethod
    def plot_fund_ret_each_year(turnover_df:pd.DataFrame,
                                mdd: float,
                                annual_ret: float):
        annual_ret = round(annual_ret, 3)
        mdd = round(mdd, 3)
        ret_year = round(turnover_df.year_ret.mean(), 3)
        p = turnover_df.set_index('year')[['year_mdd','year_ret']].plot.bar(figsize=(16,12))
        t = plt.title('return and mdd each year', fontsize=20)
        s = plt.suptitle(f'annual ret: {annual_ret}, ret mean {ret_year},mdd {mdd}', y=0.87, fontsize=15)
        h = plt.axhline(y=annual_ret, linestyle=':', label='annual ret', color='darkorange')
        r = plt.xticks(rotation=0)
        l = plt.legend(fontsize=13 ,loc='upper left')

    @staticmethod
    def plot_turnover_rate_each_year(turnover_df:pd.DataFrame, turnover_rate_yearly_avg:float):
        p = turnover_df.set_index('year')[['turnover_rate_yearly']].plot.bar(figsize=(16,12))
        t = plt.title('turnover rate each year %', fontsize=20)
        s = plt.suptitle(f'mean {turnover_rate_yearly_avg}', y=0.87, fontsize=15)
        y = turnover_df['turnover_rate_yearly'].mean()
        h = plt.axhline(y=y, linestyle=':', label='turnover rate mean')
        r = plt.xticks(rotation=0)
        l = plt.legend(fontsize=13 ,loc='upper left')

    @staticmethod
    def plot_fund_score(fund_mv:pd.DataFrame, 
                        fund_weights_history:dict, 
                        trade_history:dict,
                        index_price:pd.DataFrame,
                        asset_weights:dict,
                        fund_info:pd.DataFrame,
                        fund_nav:pd.DataFrame,
                        fund_score:dict,
                        fund_score_raw:dict,
                        fund_indicator:pd.DataFrame,
                        asset=str,
                        is_tuning=bool,
                        ):
        fund_w = fund_weights_history
        end_date = fund_mv.index.tolist()[-1]
        date_list = list(trade_history.keys()) + [end_date]
        fund_asset_df = fund_info[['fund_id','index_id']].set_index('fund_id')
        fund_indicator = fund_indicator.pivot_table(index = ['fund_id','datetime'])
        traded_to_submit_date = {}
        for d in trade_history:
            dic = trade_history[d][0]
            traded_to_submit_date[dic.trade_date] = dic.submit_date
        res = []
        for k,v in asset_weights.items():
            v['date'] = k
            res.append(v)
        weights_df = pd.DataFrame(res).set_index('date')
        weights_df = weights_df.drop(['cash'], axis = 1).dropna()[1:]
        res = []
        for dic, s in zip(weights_df.to_dict('records'), weights_df.sum(axis = 1).values):
            res.append({ k: v/ s for k, v in dic.items()})
        df = pd.DataFrame(res)
        cols = df.columns.tolist()
        name_dic = fund_info[['fund_id','desc_name']].set_index('fund_id')
        b_d = fund_nav.index[0]
        e_d = fund_nav.index[-1]
        bench_df = index_price.loc[b_d:e_d,[asset]]
        bench_df = bench_df/bench_df.iloc[0]
        for i in range(len(date_list) - 1):   
            b_d = date_list[i]
            e_d = date_list[i+1]
            if b_d == e_d:
                break
            bench_df_tmp = bench_df.loc[b_d:e_d,:]
            f_l = [k  for k ,v in fund_w[b_d].items() if isinstance(v, float) and round(v ,3 ) > 0]
            if 'cash' in f_l:
                f_l.remove('cash')
            f_l = [f for f in f_l if fund_asset_df.loc[f,'index_id'] == asset]
            mv_b = bench_df.loc[b_d,asset]
            fund_tmp = fund_nav.loc[b_d:e_d,f_l].copy()
            fund_tmp = fund_tmp/fund_tmp.iloc[0]
            fund_tmp = fund_tmp*mv_b
            res = []
            submit_d = traded_to_submit_date[b_d]
            if len(f_l) < 1:
                continue
            fig, ax = plt.subplots(figsize= [16,12])
            plt.plot(bench_df_tmp.index, bench_df_tmp[asset], label=asset,linewidth=5.0)
            for f in f_l:
                desc = name_dic.loc[f,'desc_name']
                f_i_dict= fund_indicator.loc[f,submit_d].to_dict()
                dic = {
                    'fund_id' : f,
                    'desc_name' : desc,
                    'score' : round(fund_score[submit_d][asset][f],4),
                    'weight': round(fund_w[b_d][f], 4),
                    'score_raw': round(fund_score_raw[submit_d][asset][f],4),
                }
                for s in ['alpha','beta','fee_rate','track_err']:
                    dic[s] = round(f_i_dict[s],4)
                res.append(dic)    
                plt.plot(fund_tmp.index, fund_tmp[f], label=f+'_'+desc,linestyle='--',linewidth=3.0)    
            
            if is_tuning:
                fund_not_select = [f for f, v in fund_score[submit_d][asset].items() if f not in f_l][:10]
                if len(fund_not_select) < 1:
                    break
                fund_tmp = fund_nav.loc[b_d:e_d,fund_not_select].copy()
                fund_tmp = fund_tmp/fund_tmp.iloc[0]
                fund_tmp = fund_tmp*mv_b
                for f in fund_not_select:
                    desc = name_dic.loc[f,'desc_name']
                    f_i_dict= fund_indicator.loc[f,submit_d].to_dict()
                    dic = {
                        'fund_id' : f,
                        'desc_name' : desc,
                        'score' : round(fund_score[submit_d][asset][f],4),
                        'weight': 0,
                        'score_raw': round(fund_score_raw[submit_d][asset][f],4),
                    }
                    for s in ['alpha','beta','fee_rate','track_err']:
                        dic[s] = round(f_i_dict[s],4)
                    res.append(dic)   
                    plt.plot(fund_tmp.index, fund_tmp[f], label=f+'_'+desc,linestyle=':',linewidth=3.0)    

            plt.legend(fontsize=15 ,loc = 'lower left')
            plt.title(f'{asset} {b_d} {e_d}', fontsize=25)
            plt.suptitle(FundScoreManager().funcs[asset].__dict__, y=0.87, fontsize=18)
            ax.xaxis.set_ticks_position('top')
            fund_df = pd.DataFrame(res)
            fund_df = fund_df[['desc_name','fund_id','weight','alpha','beta','track_err','fee_rate','score_raw','score']]
            fund_df = fund_df.sort_values('score', ascending = False)
            t = plt.table(
                cellText=fund_df.values.tolist(),
                colLabels=fund_df.columns,
                loc='bottom',
                colWidths= [0.25,0.12,0.09,0.09,0.09,0.09,0.09,0.09,0.09]          
            )
            t.auto_set_font_size(False)
            t.set_fontsize(15)
            t.auto_set_column_width('fund_id')
            t.AXESPAD = 0.1
            t.scale(1, 4)
            plt.show()


import time
import datetime
import pandas as pd
from pprint import pprint 
import operator
import traceback
from .backtest import FundBacktestEngine
from .trader import AssetTrader
from ...data.manager.manager_fund import FundDataManager
from ...data.struct import AssetTrade, FundTrade, FundScoreParam, TAAParam, TaaTunerParam, AssetTradeParam, AssetWeight
from ...data.api.derived import DerivedDataApi

class TaaTuner:

    BACKTEST_NUM = 3
    INDEX_WEIGHT = 0.1
    CASH_WEIGHT = 0.9
    TAA_LOW_TYPE = 'taa_low_only'
    TAA_UP_TYPE = 'taa_up_only'

    def __init__(self, start_time:str='20050101', end_time:str='20200501'):
        self.dm = FundDataManager(start_time=start_time, end_time=end_time)
        self.start_date = datetime.datetime.strptime(start_time, '%Y%m%d').date()
        self.end_date = datetime.datetime.strptime(end_time, '%Y%m%d').date()
        self.ap = AssetTradeParam()
        self.at = AssetTrader(self.ap)

    def init(self, taa_tuner:TaaTunerParam ):
        self.dm.init(score_pre_calc=False)
        self.taa_tuner = taa_tuner
        self.saa = AssetWeight(**{'cash':self.CASH_WEIGHT, self.taa_tuner.index_id:self.INDEX_WEIGHT})
        self.taa_type = self.TAA_LOW_TYPE if self.taa_tuner.IsTaaLowOnly else self.TAA_UP_TYPE
        self.bk_each_time = None

    def grid_search(self):
        self.run_saa()
        self.test_bk_time()
        self.search_bk()
        self.first_round_analysis()
        self.search_bk_2()
        self.second_round_analysis()

    def run_saa(self):
        saa_bk = FundBacktestEngine(data_manager=self.dm, trader=self.at, taa_params=None)
        saa_bk.init()
        saa_bk.run(saa=self.saa, start_date=self.start_date, end_date=self.end_date)
        self.saa_result = saa_bk._report_helper.get_asset_stat()
        
    def search_bk(self, is_test:bool=False, bk_list=None):
        if bk_list is None:
            if is_test:
                taa_param_list = self.taa_tuner.param_list[:self.BACKTEST_NUM]
            else:
                taa_param_list = self.taa_tuner.param_list
        else:
            taa_param_list = bk_list

        self.result = []
        c = 1
        bk_num = len(taa_param_list)
        if self.bk_each_time is not None:
            cost_min = len(taa_param_list) * self.bk_each_time / 60
            print(f'totally bk {bk_num} tasks index_id: {self.taa_tuner.index_id} may cost {cost_min} mins')
        else:
            print(f'totally bk {bk_num} tasks index_id: {self.taa_tuner.index_id}')
        for taa_i in taa_param_list:
            t_0 = time.time()
            taa_bk_i = FundBacktestEngine(data_manager=self.dm, trader=self.at, taa_params=taa_i)
            taa_bk_i.init()
            taa_bk_i.run(saa=self.saa, start_date=self.start_date, end_date=self.end_date)
            taa_result = taa_bk_i._report_helper.get_asset_stat()
            dic = taa_i.__dict__
            dic['taa_ret'] = taa_result['annual_ret']
            self.result.append(dic)
            t_1 = time.time()
            print(f'finish task {c}, cost {t_1 - t_0} sec')
            c += 1
        self.result = pd.DataFrame(self.result)
        self.result['saa_annual_ret'] = self.saa_result['annual_ret']
        self.result['taa_saa_annual_ret'] = self.result['taa_ret'] - self.saa_result['annual_ret']
        self.result['taa_type'] = self.taa_type
        self.result = self.result.sort_values('taa_saa_annual_ret', ascending=False)
        print()

    def test_bk_time(self):
        print()
        print('test backtest')
        t_0 = time.time()
        self.search_bk(is_test=True)
        t_1 = time.time()
        cost_time = t_1 - t_0
        self.bk_each_time = cost_time / self.BACKTEST_NUM
        print(f'bk {self.BACKTEST_NUM} times total cost {cost_time} sec, each bk {self.bk_each_time} sec')
        t_1 = time.time()
        task_num = len(self.taa_tuner.param_list)
        print(f'first round backtest may cost {task_num * self.bk_each_time / 60} mins')
        print()

    def first_round_analysis(self): 
        self.result.to_csv(f'{self.taa_tuner.index_id}_{self.taa_type}_round1.csv')
        self.first_round_result = self.result.iloc[0]
        print('first round filter result')
        pprint(self.first_round_result)
        print()

    def search_bk_2(self):
        d = self.first_round_result 
       
        self.bk_params = []
        if d.taa_type == self.TAA_UP_TYPE:
            LowThreshold = d.LowThreshold
            LowStop = d.LowStop
            LowPlus = d.LowPlus
            for HighMinus in [_ for _ in range( int(d.HighMinus * 100) - 1,int(d.HighMinus * 100) + 2, 1)]:
                for HighStop in [_ for _ in range( int(d.HighStop * 100) - 2,int(d.HighStop * 100) + 3, 1)]:
                    for HighThreshold  in [_ for _ in range( int(d.HighThreshold * 100) - 2,int(d.HighThreshold * 100) + 3, 1)]:
                        t = TAAParam()
                        t.HighThreshold = HighThreshold / 100
                        t.HighStop = HighStop / 100
                        t.HighMinus = HighMinus / 100
                        t.LowStop = LowStop 
                        t.LowThreshold = LowThreshold 
                        t.LowPlus = LowPlus 
                        t.TuneCash = True
                        self.bk_params.append(t)
        else:
            HighThreshold = d.HighThreshold
            HighStop = d.HighStop
            HighMinus = d.HighMinus
            for LowPlus in [_ for _ in range( int(d.LowPlus * 100) - 1,int(d.LowPlus * 100) + 2, 1)]:
                for LowStop in [_ for _ in range( int(d.LowStop * 100) - 2,int(d.LowStop * 100) + 3, 1)]:
                    for LowThreshold in [_ for _ in range( int(d.LowThreshold * 100) - 2,int(d.LowThreshold * 100) + 3, 1)]:
                        t = TAAParam()
                        t.HighThreshold = HighThreshold
                        t.HighStop = HighStop
                        t.HighMinus = HighMinus
                        t.LowStop = LowStop / 100
                        t.LowThreshold = LowThreshold / 100
                        t.LowPlus = LowPlus / 100
                        t.TuneCash = True
                        self.bk_params.append(t)
        self.search_bk(is_test=False, bk_list=self.bk_params)
        
    def second_round_analysis(self):
        self.result.to_csv(f'{self.taa_tuner.index_id}_{self.taa_type}_round2.csv')
        self.second_round_result = self.result.iloc[0]
        print('second round filter result')
        pprint(self.second_round_result)


class SaaTuner:

    ASSET_LIST = ['hs300', 'csi500', 'gem', 'sp500rmb', 'national_debt', 'gold']
    ROUND1_STEP = 5
    ROUND2_STEP = 1
    ROUND1_CSV = 'saa_tuner_round_1.csv'
    ROUND2_CSV = 'saa_tuner_round_2.csv'
    FINAL_CSV = 'saa_tuner_round_final.csv'

    def __init__(self, start_time:str='20050101', end_time:str='20200520'):
        self.dm = FundDataManager(start_time=start_time, end_time=end_time)
        self.start_date = datetime.datetime.strptime(start_time, '%Y%m%d').date()
        self.end_date = datetime.datetime.strptime(end_time, '%Y%m%d').date()
        self.derived = DerivedDataApi()

    def init(self):
        self.result_round1 = []
        self.result_round2 = []
        self.result_target = []
        self.dm.init(score_pre_calc=False)
        self.get_ports_saa()
        self.test_bk_time()
        
    def tune(self):
        self.first_round()
        self.second_round()
        self.final_round()

    def formal_print(self, sentence):
        print(sentence)
        print()

    def get_ports_saa(self):
        asset_info = self.derived.get_asset_allocation_info()
        asset_info = asset_info[['allocation_id','hs300','csi500','gem','sp500rmb','national_debt','gold','cash']].set_index('allocation_id')
        self.port_saa_dic = asset_info.to_dict('index')

    def bk_unit(self, asset:str, amt_diff:float):
        saa = AssetWeight(**{asset:10/100,'cash':90/100})
        asset_param = AssetTradeParam(EnableCommission=True, PurchaseDiscount=0.15, MinActionAmtDiff=amt_diff) 
        t = AssetTrader(asset_param)
        b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
        b.init()
        b.run(saa=saa,start_date=self.start_date, end_date=self.end_date)
        res = b.get_asset_result()
        return {'mdd': res['mdd'], 
                'annual_ret': res['annual_ret'], 
                'ret_over_mdd': res['ret_over_mdd'], 
                'asset': asset, 
                'amt_diff': amt_diff, 
                'start_date': self.start_date, 
                'end_date': self.end_date}

    def bk_unit_target(self, amt_diff:float):
        saa = AssetWeight(    
            hs300=15/100,
            csi500=5/100,
            gem=3/100,
            sp500rmb=7/100,
            national_debt=60/100,
            gold=7/100,
            cash=3/100
        )  
        asset_param = AssetTradeParam(EnableCommission=True, PurchaseDiscount=0.15, MinActionAmtDiff=amt_diff) 
        t = AssetTrader(asset_param)
        b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
        b.init()
        b.run(saa=saa,start_date=self.start_date, end_date=self.end_date)
        res = b.get_asset_result()
        return {'mdd': res['mdd'], 
                'annual_ret': res['annual_ret'], 
                'ret_over_mdd': res['ret_over_mdd'], 
                'amt_diff': amt_diff, 
                'start_date': self.start_date, 
                'end_date': self.end_date}
    
    def bk_input_saa_amt_diff(self, saa_dict:dict, amt_diff:float, port_id:int=0):
        try:
            saa = saa_dict#AssetWeight(saa_dict)  
            asset_param = AssetTradeParam(EnableCommission=True, PurchaseDiscount=0.15, MinActionAmtDiff=amt_diff) 
            t = AssetTrader(asset_param)
            b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
            b.init()
            b.run(saa=saa,start_date=self.start_date, end_date=self.end_date)
            res = b.get_asset_result()
            res['MinActionAmtDiff'] = amt_diff
            res['start_date'] = self.start_date
            res['end_date'] = self.end_date
            res['port_id'] = port_id
            return res
        except Exception as e:
            print(e)
            traceback.print_stack()
            print('boom')
            print(f'port_id {port_id} amt_diff {amt_diff}')
            print(saa_dict)    
            return None
        
    def test_bk_time(self):
        self.formal_print('test backtest')
        t_0 = time.time()
        asset = self.ASSET_LIST[0]
        amt_diff = [i/1000 for i in range(30,40,2)]
        for amt_diff_i in amt_diff:
            self.bk_unit(asset, amt_diff_i)
        t_1 = time.time()
        cost_time = t_1 - t_0
        self.bk_each_time = cost_time / len(amt_diff)
        self.formal_print(f'bk {len(amt_diff)} times total cost {cost_time} sec, each bk {self.bk_each_time} sec')
        
    def first_round(self):
        loop_layer_1 = self.ASSET_LIST
        loop_layer_2 = [_/10000 for _ in range(100, 905, self.ROUND1_STEP)]
        predict_time = len(loop_layer_1) * len(loop_layer_2) * self.bk_each_time / 60
        self.formal_print(f'###first round bk may cost {predict_time} mins')
        bk_time0 = time.time()
        for asset in loop_layer_1:
            asset_t0 = time.time()
            for amt_diff_i in loop_layer_2:
                self.result_round1.append(self.bk_unit(asset, amt_diff_i))
            asset_t1 = time.time()
            cost_m = (asset_t1 - asset_t0) / 60
            self.formal_print(f' - asset {asset} bk finish cost {cost_m} mins')
        bk_time1 = time.time()
        cost_m = (bk_time1 - bk_time0) / 60
        self.formal_print(f'first round finish cost cost {cost_m} mins')
        pd.DataFrame(self.result_round1).to_csv(self.ROUND1_CSV)

    def second_round(self):
        self.round1_df = pd.read_csv(self.ROUND1_CSV, index_col=0)
        loop_layer_1 = self.ASSET_LIST
        layer_2_lens = len([ i for i in range(-self.ROUND1_STEP+1, self.ROUND1_STEP, self.ROUND2_STEP)])
        predict_time = len(loop_layer_1) * layer_2_lens * self.bk_each_time / 60
        self.formal_print(f'###second round bk may cost {predict_time} mins')
        bk_time0 = time.time()
        for asset in loop_layer_1:
            amt_diff_0 = self.round1_df[self.round1_df.asset == asset].sort_values('ret_over_mdd', ascending=False).amt_diff.values[0]
            amt_diff_int = int(amt_diff_0 * 10000)
            for amt_diff_i in range(amt_diff_int - self.ROUND1_STEP + 1, amt_diff_int + self.ROUND1_STEP, self.ROUND2_STEP):
                amt_diff_i = amt_diff_i / 10000
                self.result_round2.append(self.bk_unit(asset, amt_diff_i))
        bk_time1 = time.time()
        cost_m = (bk_time1 - bk_time0) / 60
        self.formal_print(f'second round finish cost cost {cost_m} mins')
        pd.DataFrame(self.result_round2).to_csv(self.ROUND2_CSV)

    def final_round(self):
        self.round2_df = pd.read_csv(self.ROUND2_CSV, index_col=0)
        res = []
        for asset in self.ASSET_LIST:
            res_i = self.round2_df[self.round2_df.asset == asset].sort_values('ret_over_mdd', ascending=False).reset_index(drop=True).iloc[0]
            res.append(res_i)
        self.final_df = pd.DataFrame(res)
        self.formal_print('###final result sort by ret over mdd')
        print(self.final_df)
        self.final_df.to_csv(self.FINAL_CSV)
    
    def target_backtest(self):
        loop_layer = [_/10000 for _ in range(100, 905, self.ROUND1_STEP)]
        for amt_diff_i in loop_layer:
            self.result_target.append(self.bk_unit_target(amt_diff_i))
        self.target_df = pd.DataFrame(self.result_target)
        res = self.target_df.sort_values('ret_over_mdd', ascending=False).reset_index(drop=True).head()
        print(res)

    def bk_fee_search(self, fee_con:bool, MinActionAmtDiff:float):
        saa = AssetWeight(    
            hs300=15/100,
            csi500=5/100,
            gem=3/100,
            sp500rmb=7/100,
            national_debt=60/100,
            gold=7/100,
            cash=3/100
        )  
        asset_param = AssetTradeParam(EnableCommission=fee_con, PurchaseDiscount=0.15, MinActionAmtDiff=MinActionAmtDiff) 
        t = AssetTrader(asset_param)
        b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
        b.init()
        b.run(saa=saa,start_date=self.start_date, end_date=self.end_date)
        res = b.get_asset_result()
        res['MinActionAmtDiff'] = MinActionAmtDiff
        res['fee_status'] = fee_con
        res['start_date'] = self.start_date
        res['end_date'] = self.end_date
        return res

    def search_fee(self):
        self.result_fee = []
        #for fee_con in [True, False]:
        fee_con = False
        for MinActionAmtDiff in [0.02,0.03,0.04,0.05,0.06,0.065,0.07,0.08]:
            res_i = self.bk_fee_search(fee_con, MinActionAmtDiff)
            self.result_fee.append(res_i)
        #self.result_fee_df = pd.DataFrame(self.result_fee)

    def bk_unit_double_asset(self, saa, amt_dif):
        asset_param = AssetTradeParam(EnableCommission=True, PurchaseDiscount=0.15, MinActionAmtDiff=amt_dif) 
        t = AssetTrader(asset_param)
        b = FundBacktestEngine(data_manager=self.dm, trader=t, taa_params=None)
        b.init()
        b.run(saa=saa)
        res = b.get_asset_result()
        res['amt_dif'] = amt_dif
        res['rebalance_details'] = b._trader.rebalance_details
        return res

    def bk_loop_selected_amt_diff(self, asset_1, asset_2):
        result = []
        for w in range(10,91,2):
            saa = AssetWeight(**{asset_1:w/100,asset_2:(99-w)/100, 'cash':1/100})
            amt_dif = 50 / 1000
            res_i = self.bk_unit_double_asset(saa=saa, amt_dif=amt_dif)
            result.append(res_i)
        return pd.DataFrame(result)

    def find_trigger_asset(self, dic):
        res = []
        for d, k in dic.items():
            if k:
                res.append({'date':d,'trigger_asset':max(k.items(), key=operator.itemgetter(1))[0]})
        trigger_list = pd.DataFrame(res).trigger_asset.tolist()
        trigger_dic = {i: trigger_list.count(i) for i in set(trigger_list)}
        trigger_asset = max(trigger_dic.items(), key=operator.itemgetter(1))[0]
        return trigger_asset

    def find_trigger_process(self, asset_1, asset_2):
        result = self.bk_loop_selected_amt_diff(asset_1,asset_2)
        res = []
        for idx, r in result.iterrows():
            if len(r.rebalance_date) == 2:
                res.append('unvalid bk, rebalence time <= 2')
                continue
            if idx == 0:
                res.append('None')
                continue
            trigger_asset=self.find_trigger_asset(r.rebalance_details)
            res.append(trigger_asset)
        result['trigger_asset'] = res
        result['rebalance_times'] = [len(_) for _ in result['rebalance_date']]
        return result

if __name__ == "__main__":
    
    tt = TaaTunerParam()
    tt.IsTaaUpOnly = True
    tt.index_id = 'hs300'

    LowStop = -100
    LowThreshold = -100
    LowPlus = 0

    tt.param_list = []
    for HighThreshold in range(80, 97, 2):
        for HighStop in range(50, 71, 2):
            for HighMinus in [3 , 5, 7]:
                t = TAAParam()
                t.HighThreshold = HighThreshold / 100
                t.HighStop = HighStop / 100
                t.HighMinus = HighMinus / 100
                t.LowStop = LowStop / 100
                t.LowThreshold = LowThreshold / 100
                t.LowPlus = LowPlus / 100
                t.TuneCash = True
                tt.param_list.append(t)

    bk = TaaTuner()
    bk.init(tt)   
    bk.grid_search()

    '''
    asset_1 = 'hs300'
    asset_2 = 'csi500'
    hs300_csi500 = find_trigger_process(asset_1, asset_2)
    hs300_csi500[[asset_1,asset_2,'trigger_asset','rebalance_times']]
    '''


from ..struct import FundScoreParam
from .data_tables import FundDataTables
import dataclasses
import pandas as pd 
import re

@dataclasses.dataclass
class ScoreFunc:

    alpha: float=0 # weight of alpha
    beta: float=0  # weight of abs(1-beta), or beta's deviation from 1
    track_err: float=0 # weight of track_err
    fee_rate: float=0 # weight of fee_rate
    ir: float=0 # weight of ir

    def get(self, data):
        return self.alpha * data.alpha + self.beta * abs(1-data.beta) + self.track_err * data.track_err + self.fee_rate * data.fee_rate + self.ir * data.ir


class FundScoreManager:

    def __init__(self):
        self.params = None
        self.dts = None
        self.funcs = {
            'hs300': ScoreFunc(alpha=0.3, beta=-0.6, fee_rate=-0.1),
            'csi500': ScoreFunc(alpha=0.8, beta=-0.2),
            'gem': ScoreFunc(alpha=0.2, beta=-0.4, track_err=-0.4),

            'national_debt': ScoreFunc(alpha=0.6, fee_rate=-0.2, track_err=-0.2),
            'mmf':ScoreFunc(alpha=0.6, fee_rate=-0.3, track_err=-0.1),
            'credit_debt': ScoreFunc(alpha=0.2, fee_rate=-0.2, track_err=-0.8),

            'sp500rmb': ScoreFunc(fee_rate=-0.2, track_err=-0.8),
            'dax30rmb': ScoreFunc(fee_rate=-0.2, track_err=-0.8),
            'n225rmb': ScoreFunc(fee_rate=-0.2, track_err=-0.8),

            'gold': ScoreFunc(alpha=0.2, fee_rate=-0.2, track_err=-0.8),
            'oil': ScoreFunc(alpha=0.2, fee_rate=-0.2, track_err=-0.8),
            'real_state': ScoreFunc(alpha=0.2, fee_rate=-0.2, track_err=-0.8),

        }
        self.score_cache = None
        self.score_raw_cache = None

    def set_param(self, score_param: FundScoreParam):
        self.params = score_param

    def set_dts(self, dts: FundDataTables):
        self.dts = dts

    def get_fund_score(self, dt, index_id) -> dict:
        func = self.funcs.get(index_id)
        assert self.params and self.dts, 'cannot provide fund_score without params or data tables'
        try:
            cur_d = self.dts.index_fund_indicator_pack.loc[index_id, dt]
        except:
            return {}, {}
        if len(cur_d) == 0:
            return {}, {}
        elif len(cur_d) == 1:
            return {cur_d.index[0]: 1}, {cur_d.index[0]: 1}
        else:
            assert func, f'score function does not implemented {index_id}'
            cur_d_sta = cur_d.apply(lambda x: (x - x.mean() + 1e-6)/ (x.std() + 1e-6), axis=0)
            cur_d_sta['beta'] = cur_d['beta']
            score_raw = cur_d_sta.apply(lambda x: func.get(x), axis=1)
            score = (score_raw - score_raw.min()) / (score_raw.max() - score_raw.min())
            return { fund_id: s for fund_id, s in score.iteritems() }, { fund_id: s for fund_id, s in score_raw.iteritems() }

    def get_fund_scores(self, dt, index_list) -> dict:
        if not self.score_cache:
            score = {}
            score_raw = {}
            for index_id in index_list:
                score[index_id], score_raw[index_id] = self.get_fund_score(dt, index_id)
            return score, score_raw
        return self.score_cache.get(dt, {}), self.score_raw_cache.get(dt, {})

    def pre_calculate(self) -> dict:
        self.score_cache = {}
        self.score_raw_cache = {}
        pad = self.dts.fund_indicator.pivot_table(index=['datetime', 'index_id'])
        for dt, index_id in pad.index:
            if not dt in self.score_cache:
                self.score_cache[dt] = {}
                self.score_raw_cache[dt] = {}
            cur_d = self.dts.index_fund_indicator_pack.loc[index_id, dt]
            if cur_d.shape[0] > 1:
                cur_d_sta = cur_d.apply(lambda x: (x - x.mean() + 1e-6)/ (x.std() + 1e-6), axis=0)
            else:
                cur_d_sta = cur_d
            cur_d_sta['beta'] = cur_d['beta']
            func = self.funcs.get(index_id)
            score_raw = cur_d_sta.apply(lambda x: func.get(x), axis=1)
            score = (score_raw - score_raw.min()) / (score_raw.max() - score_raw.min())
            if score.shape[0] == 1:
                score.values[0] = 1
            self.score_cache[dt][index_id] = { fund_id: s for fund_id, s in score.iteritems() }
            self.score_raw_cache[dt][index_id] = { fund_id: s for fund_id, s in score_raw.iteritems() }

    def filter_score(self, score_dict, fund_id_to_name, fund_name_to_id):
        # 同一天同资产 同基金A B 在 选B
        # 同一天同资产 同基金A C 在 选C
        check_char_list = ['B','C']
        for check_char in check_char_list:
            for d, score_d in score_dict.items():
                for index_id, score_index in score_d.items():
                    index_fund_name = [fund_id_to_name[_] for _ in list(score_index.keys())]
                    c_list = [word for word in index_fund_name if check_char in word]
                    # 如果d日, index_id资产下含有C类基金（包含假C类 如名字含有MSCI）
                    if len(c_list) > 0:
                        for c_fund in c_list:
                            # 替换基金名字里最后一个C 变为A， 这样可以处理假C类基金
                            a_fund = re.sub(f'{check_char}$', 'A', c_fund)
                            # 如果A类有评分， 赋值0
                            if a_fund in index_fund_name:
                                a_fund_id = fund_name_to_id[a_fund]
                                score_dict[d][index_id][a_fund_id] = 0
        return  score_dict

    def remove_a_score_if_c_exsit(self, fund_info:pd.DataFrame):
        fund_id_to_name = fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        fund_name_to_id = fund_info[['fund_id','desc_name']].set_index('desc_name').to_dict()['fund_id']
        self.score_cache = self.filter_score(self.score_cache, fund_id_to_name, fund_name_to_id)
        self.score_raw_cache = self.filter_score(self.score_raw_cache, fund_id_to_name, fund_name_to_id)


#!/usr/bin/env python
# coding: utf-8
import numpy as np
import pandas as pd
import rqdatac as rq
import pickle as pkl
from typing import List
from .EMQuantAPI_Python.python3.EmQuantAPI import *
import datetime
import json
import os
import csv
import traceback
from sqlalchemy import distinct
from sqlalchemy.orm import sessionmaker
from ...wrapper.mysql import RawDatabaseConnector
from ...view.raw_models import *
from ...api.basic import BasicDataApi
from ...api.raw import RawDataApi

class EmRawDataDownloader(object):

    def __init__(self):
        # ForceLogin
        # 取值0，当线上已存在该账户时，不强制登录
        # 取值1，当线上已存在该账户时，强制登录，将前一位在线用户踢下线
        options = 'ForceLogin=1'
        loginResult = c.start(options, mainCallBack=self._main_callback)
        if(loginResult.ErrorCode != 0):
            print('EM login failed')
            exit()

        self._scale_change_dates = ['0331', '0630', '0930', '1231']

        self.updated_count = {}

    def _upload_raw(self, df: pd.DataFrame, table_name: str):
        print(table_name)
        df.to_sql(table_name, RawDatabaseConnector().get_engine(), index=False, if_exists='append')

        if table_name not in self.updated_count:
            self.updated_count[table_name] = 0
        self.updated_count[table_name] += df.shape[0]

    def _main_callback(self, quantdata):
        '''
        mainCallback 是主回调函数，可捕捉连接错误
        该函数只有一个为c.EmQuantData类型的参数quantdata
        :param quantdata: c.EmQuantData
        '''
        print(f'MainCallback: {quantdata}')

    def _get_date_list(self, start_date, end_date):
        start_datetime = datetime.datetime.strptime(start_date, '%Y%m%d')
        end_datetime = datetime.datetime.strptime(end_date, '%Y%m%d')
        date_list = []
        while start_datetime <= end_datetime:
            date_list.append(start_datetime.strftime('%Y%m%d'))
            start_datetime += datetime.timedelta(days=1)
        return date_list

    def em_fund_nav_history(self, nav_file_dir):
        try:
            index = 1
            for filename in os.listdir(nav_file_dir):
                if filename.endswith(".xls"):
                    print(f'{index}: {filename}')
                    index += 1
                    fund_nav = pd.concat(pd.read_excel(os.path.join(nav_file_dir, filename), sheet_name=None),
                                        ignore_index=True)
                    fund_nav = fund_nav.drop(['简称'], axis=1)
                    fund_nav = fund_nav.rename(columns={
                        '代码': 'CODES',
                        '时间': 'DATES',
                        '单位净值(元)': 'ORIGINALUNIT',
                        '累计净值(元)': 'ORIGINALNAVACCUM',
                        '复权净值(元)': 'ADJUSTEDNAV',
                        '万份基金单位收益(元)': 'UNITYIELD10K',
                        '7日年化收益率(%)': 'YIELDOF7DAYS'
                    })

                    fund_nav = fund_nav[(fund_nav['DATES'] != '--')
                                        & (~fund_nav['DATES'].isnull())]
                    fund_nav = fund_nav.replace('--', np.nan)
                    self._upload_raw(fund_nav, EmFundNav.__table__.name)
                else:
                    continue
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def em_index_price_history(self, em_ids, start_date, end_date):
        em_id_str = ','.join(em_ids) if em_ids is list else em_ids
        df = c.csd(em_id_str, "CLOSE", start_date, end_date,
                   "period=1,adjustflag=1,curtype=1,order=1,market=CNSESH,Ispandas=1")

        df = df.reset_index().rename(columns={
            'CODES': 'em_id',
            'DATES': 'datetime',
            'CLOSE': 'close'
        })

        self._upload_raw(df, EmIndexPrice.__table__.name)

    def em_fund_scale_history(self):
        date_funds = {}
        with RawDatabaseConnector().managed_session() as db_session:
            try:
                dates = db_session.query(
                    distinct(EmFundNav.DATES)
                ).all()

                dates = [d for d, in dates]
                dates.sort()
                start_date = dates[0]
                for i in range(1, 100):
                    dates.insert(0, start_date - datetime.timedelta(days=i))

                for d in dates:
                    month_day = d.strftime('%m%d')
                    if month_day not in self._scale_change_dates:
                        continue

                    print(d)

                    fund_ids = db_session.query(
                        distinct(EmFundNav.CODES)
                    ).filter(
                        EmFundNav.DATES >= d,
                        EmFundNav.DATES <= d + datetime.timedelta(days=100)
                    ).all()

                    fund_ids = [f for f, in fund_ids]
                    fund_id_str = ','.join(fund_ids)

                    df = c.css(fund_id_str, "FUNDSCALE",
                               f"EndDate={d},Ispandas=1").reset_index()
                    df['DATES'] = d

                    self._upload_raw(df, EmFundScale.__table__.name)
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))
                return None

    def em_fund_nav(self, start_date, end_date):
        try:
            # Get all fund ids, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            fund_id_result = c.sector('507013', end_date)
            if fund_id_result.ErrorCode != 0:
                print(f'Failed to get fund id list: {fund_id_result.ErrorMsg}')
                return

            fund_id_list = fund_id_result.Data[0::2]
            fund_id_str = ','.join(fund_id_list)

            df = c.csd(fund_id_str, 'ORIGINALUNIT,ORIGINALNAVACCUM,ADJUSTEDNAV,10KUNITYIELD,YIELDOF7DAYS',
                    start_date, end_date, 'AdjustFlag=2,Ispandas=1')

            # print(df)
            df.reset_index(inplace=True)
            df['UNITYIELD10K'] = df['10KUNITYIELD']
            df = df.drop(['10KUNITYIELD'], axis=1)
            self._upload_raw(df, EmFundNav.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def em_fund_scale(self, start_date, end_date):
        try:
            # Get all fund ids, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            fund_id_result = c.sector('507013', end_date)
            if fund_id_result.ErrorCode != 0:
                print(f'Failed to get fund id list: {fund_id_result.ErrorMsg}')
                return

            fund_id_list = fund_id_result.Data[0::2]
            fund_id_str = ','.join(fund_id_list)

            for d in self._get_date_list(start_date, end_date):
                df = c.css(fund_id_str, 'FUNDSCALE', f'EndDate={d},Ispandas=1').reset_index()
                df['DATES'] = d
                self._upload_raw(df, EmFundScale.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def em_index_price(self, em_ids, start_date, end_date):
        start_date = datetime.datetime.strptime(start_date, '%Y%m%d') - datetime.timedelta(days=5)
        start_date = start_date.strftime('%Y%m%d')
        self.delete_em_index_price(em_ids, start_date, end_date)
        try:
            res = []
            if isinstance(em_ids, list):
                for em_id in em_ids:
                    df = c.csd(em_id, "CLOSE", start_date, end_date,"period=1,adjustflag=1,curtype=1,order=1,market=CNSESH,Ispandas=1")
                    res.append(df)
                df = pd.concat(res)
            else:
                df = c.csd(em_ids, "CLOSE", start_date, end_date,"period=1,adjustflag=1,curtype=1,order=1,market=CNSESH,Ispandas=1")

            df = df.reset_index().rename(columns={
                'CODES': 'em_id',
                'DATES': 'datetime',
                'CLOSE': 'close'
            })

            self._upload_raw(df, EmIndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def delete_em_index_price(self, em_ids, start_date, end_date):
        try:
            RawDataApi().delete_em_index_price(em_ids, start_date, end_date)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False    

    def em_index(self, start_date, end_date):
        # ----------------------
        # H11073.CSI,信用债
        # SPX.GI,标普500
        # GDAXI.GI,德国DAX
        # N225.GI,日经225
        # 000012.SH,利率债
        # 000300.SH,沪深300
        # 000905.SH,中证500
        # 399006.SZ,创业板
        # 801181.SWI,房地产开发申万
        # ----------------------
        # USDCNY.IB,美元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # JPYCNY.IB,日元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # EURCNY.IB,欧元兑人民币市询价,TODO: '''ErrorCode=10001012, ErrorMsg=insufficient user access, Data={}'''
        # ----------------------
        # AU0.SHF,沪金主力连续,TODO: always get None
        # BC.ICE,布伦特原油当月连续,TODO: always get None
        # ----------------------
        codes = (
                 'H11073.CSI,SPX.GI,GDAXI.GI,N225.GI,000012.SH,000300.SH,000905.SH,399006.SZ,801181.SWI,'
                 'USDCNY.IB,JPYCNY.IB,EURCNY.IB,'
                #  'AU0.SHF,BC.ICE'
        )
        df = c.csd(codes, 'CLOSE', start_date, end_date, 'Ispandas=1')
        if df is not None:
            df.reset_index(inplace=True)
            print(df)
            # self._upload_raw(df, EmIndexPrice.__table__.name)

    # 从em客户端导出数据后读取并处理
    def em_stock_price_history(self, stock_price_file_dir: str):
        import os
        for index, filename in enumerate(os.listdir(stock_price_file_dir)):
            if not filename.endswith(".xls"):
                continue
            index += 1
            print(f'{index}: {filename}')
            stock_price = pd.concat(pd.read_excel(os.path.join(stock_price_file_dir, filename), sheet_name=None, na_values='--'),
                                    ignore_index=True)
            # 将列名字改为从API获取到数据时的列名字，便于后边统一处理
            stock_price = stock_price.drop(['简称'], axis=1).rename(columns={
                '代码': 'CODES',
                '时间': 'DATES',
                '开盘价(元)': 'OPEN',
                '收盘价(元)': 'CLOSE',
                '最高价(元)': 'HIGH',
                '最低价(元)': 'LOW',
                '前收盘价(元)': 'PRECLOSE',
                '均价(元)': 'AVERAGE',
                '成交量(股)': 'VOLUME',
                '成交金额(元)': 'AMOUNT',
                '换手率(%)': 'TURN',
                '交易状态': 'TRADESTATUS',
                '内盘成交量': 'BUYVOL',
                '外盘成交量': 'SELLVOL',
            })

            # 过滤掉个股未上市时的数据，以及日期为nan的行（表示整行为无效数据）
            stock_price = stock_price.loc[stock_price['TRADESTATUS'] != '未上市'].loc[~stock_price['DATES'].isna()].rename(columns=lambda x: EmStockPrice.__getattribute__(EmStockPrice, x).name)
            # 更新到db
            self._upload_raw(stock_price, EmStockPrice.__table__.name)

    def em_stock_price(self, start_date: str, end_date: str):
        try:
            # 获取区间内所有交易日
            tradedates = c.tradedates(start_date, end_date, "period=1,order=1,market=CNSESH")
            if tradedates.ErrorCode != 0:
                print(f'[em_stock_price] failed to get trade dates: {tradedates.ErrorMsg}')
                return
            if len(tradedates.Data) == 0:
                print(f'[em_stock_price] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return

            # 获取所有A股股票的ID, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            # 全部A股股票
            stock_id_result = c.sector('001004', end_date)
            if stock_id_result.ErrorCode != 0:
                print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                return

            df_list = []
            for date in tradedates.Data:
                # 获取个股股价信息
                temp_df = c.csd(stock_id_result.Codes,
                                'OPEN,CLOSE,HIGH,LOW,PRECLOSE,AVERAGE,VOLUME,AMOUNT,TURN,TRADESTATUS,TNUM,BUYVOL,SELLVOL',
                                date, date, 'period=1,adjustflag=1,curtype=1,order=1,market=0,Ispandas=1')
                if not isinstance(temp_df, pd.DataFrame) and temp_df.ErrorCode != 0:
                    print(f'failed to get stock price info: {temp_df.ErrorMsg} (date){date}')
                    return
                # 过滤掉个股未上市时的数据，以及日期为nan的行（表示整行为无效数据）
                df_list.append(temp_df.loc[temp_df['TRADESTATUS'] != '未上市'].loc[~temp_df['DATES'].isna()])

            df = pd.concat(df_list).reset_index().rename(columns=lambda x: EmStockPrice.__getattribute__(EmStockPrice, x).name)
            # 更新到db
            self._upload_raw(df, EmStockPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def em_stock_post_price(self, start_date: str, end_date: str):
        try:
            # 获取区间内所有交易日
            tradedates = c.tradedates(start_date, end_date, "period=1,order=1,market=CNSESH")
            if tradedates.ErrorCode != 0:
                print(f'[em_stock_post_price] failed to get trade dates: {tradedates.ErrorMsg}')
                return
            if len(tradedates.Data) == 0:
                print(f'[em_stock_post_price] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return

            # 获取所有A股股票的ID, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            # 全部A股股票
            stock_id_result = c.sector('001004', end_date)
            if stock_id_result.ErrorCode != 0:
                print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                return

            df_list = []
            for date in tradedates.Data:
                # 获取个股股价信息
                temp_df = c.csd(stock_id_result.Codes,
                                'OPEN,CLOSE,HIGH,LOW,PRECLOSE,AVERAGE,VOLUME,AMOUNT,TRADESTATUS,TAFACTOR',
                                date, date, 'period=1,adjustflag=2,curtype=1,order=1,market=0,Ispandas=1')
                if not isinstance(temp_df, pd.DataFrame) and temp_df.ErrorCode != 0:
                    print(f'failed to get stock price info: {temp_df.ErrorMsg} (date){date}')
                    return
                # 过滤掉个股未上市时的数据，以及日期为nan的行（表示整行为无效数据）
                df_list.append(temp_df.loc[temp_df['TRADESTATUS'] != '未上市'].loc[~temp_df['DATES'].isna()])

            df = pd.concat(df_list).reset_index().rename(columns=lambda x: EmStockPostPrice.__getattribute__(EmStockPostPrice, x).name)
            # 更新到db
            self._upload_raw(df, EmStockPostPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def em_stock_info(self, end_date: str):
        # 获取所有A股股票的ID, 功能函数-板块成分
        # http://quantapi.eastmoney.com/Cmd/Sector?from=web
        # 全部A股股票
        stock_id_result = c.sector('001004', end_date)
        if stock_id_result.ErrorCode != 0:
            print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
            return

        df_list: List[pd.DataFrame] = []
        # 获取无参数指标
        df_no_param = c.css(stock_id_result.Codes,
                            'NAME,LISTDATE,FINPURCHORNOT,FINSELLORNOT,STOHSTOCKCONNECTEDORNOT,SHENGUTONGTARGET,ENGNAME,COMPNAME,COMPNAMEENG',
                            'Ispandas=1')
        if not isinstance(df_no_param, pd.DataFrame) and df_no_param.ErrorCode != 0:
            print(f'failed to get stock info(no param): {df_no_param.ErrorMsg}')
        else:
            # 将其中'是'/'否'的结果转换为True/False
            columns = ['FINPURCHORNOT', 'FINSELLORNOT', 'STOHSTOCKCONNECTEDORNOT', 'SHENGUTONGTARGET']
            df_no_param.loc[:, columns] = df_no_param.loc[:, columns].apply(lambda x: x.apply(lambda x: True if x == '是' else False))
            df_list.append(df_no_param)

        # 获取行业分类共有三级的指标
        df_lv4 = c.css(stock_id_result.Codes, "BLEMINDCODE,BLSWSINDCODE,SW2014CODE,EMINDCODE2016,CITICCODE2020", "ClassiFication=4,Ispandas=1")
        if not isinstance(df_lv4, pd.DataFrame) and df_lv4.ErrorCode != 0:
            print(f'failed to get stock info(lv4): {df_lv4.ErrorMsg}')
        else:
            df_list.append(df_lv4)

        # 获取行业分类共有二级的指标
        df_lv3 = c.css(stock_id_result.Codes, "BLCSRCINDCODE,CSRCCODENEW", "ClassiFication=3,Ispandas=1")
        if not isinstance(df_lv3, pd.DataFrame) and df_lv3.ErrorCode != 0:
            print(f'failed to get stock info(lv3): {df_lv3.ErrorMsg}')
        else:
            df_list.append(df_lv3)

        # 获取行业分类共有四级的指标
        df_lv5 = c.css(stock_id_result.Codes, "CSINDCODE2016,GICSCODE", "ClassiFication=5,Ispandas=1")
        if not isinstance(df_lv5, pd.DataFrame) and df_lv5.ErrorCode != 0:
            print(f'failed to get stock info(lv5): {df_lv5.ErrorMsg}')
        else:
            df_list.append(df_lv5)

        if len(df_list) > 0:
            # 将每部分按列粘到一起
            df = pd.concat(df_list, axis=1)
            # 不需要存日期的列，去除索引并做列名转换
            df = df.drop(columns='DATES').reset_index().rename(columns=lambda x: EmStockInfo.__getattribute__(EmStockInfo, x).name)
            # 更新到db
            self._upload_raw(df, EmStockInfo.__table__.name)

    def add_columns_to_fin_fac(self, end_date: str):
        # 获取所有A股股票的ID, 功能函数-板块成分
        # http://quantapi.eastmoney.com/Cmd/Sector?from=web
        # 全部A股股票
        stock_id_result = c.sector('001004', end_date)
        if stock_id_result.ErrorCode != 0:
            print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
            return

        # 获取主营收入构成(按产品)
        df = c.css(stock_id_result.Codes, "MBSALESCONS", f'ReportDate={end_date},type=2,Ispandas=1')
        if not isinstance(df, pd.DataFrame) and df.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df.ErrorMsg}')
            return

        Session = sessionmaker(RawDatabaseConnector().get_engine())
        db_session = Session()
        for row in db_session.query(EmStockFinFac):
            if row.DATES != datetime.datetime.strptime(end_date, '%Y%m%d').date().year:
                continue
            if row.CODES in df.index:
                row.MBSALESCONS_P = df.at[row.CODES, 'MBSALESCONS']
        db_session.commit()
        db_session.close()

    def em_stock_fin_fac(self, end_date: str):
        # 获取所有A股股票的ID, 功能函数-板块成分
        # http://quantapi.eastmoney.com/Cmd/Sector?from=web
        # 全部A股股票
        stock_id_result = c.sector('001004', end_date)
        if stock_id_result.ErrorCode != 0:
            print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
            return

        df_list: List[pd.DataFrame] = []
        # 获取无参数指标
        df_no_param = c.css(stock_id_result.Codes,
                            'EBIT,EBITDA,EXTRAORDINARY,LOWERDIANDNI,GROSSMARGIN,OPERATEINCOME,INVESTINCOME,EBITDRIVE,TOTALCAPITAL,WORKINGCAPITAL,\
                             NETWORKINGCAPITAL,TANGIBLEASSET,RETAINED,INTERESTLIBILITY,NETLIBILITY,EXINTERESTCL,EXINTERESTNCL,FCFF,FCFE,DA,FCFFDRIVE,\
                             PERFORMANCEEXPRESSCUTPARENTNI,PERFORMANCEEXPRESSPARENTNI,MBSALESCONS,GPMARGIN,NPMARGIN,EXPENSETOOR,INVTURNRATIO,ARTURNRATIO,ROEAVG,ROEWA,EPSBASIC,EPSDILUTED,\
                             BPS,BALANCESTATEMENT_25,BALANCESTATEMENT_46,BALANCESTATEMENT_93,BALANCESTATEMENT_103,BALANCESTATEMENT_141,\
                             BALANCESTATEMENT_140,INCOMESTATEMENT_9,INCOMESTATEMENT_48,INCOMESTATEMENT_60,INCOMESTATEMENT_61,INCOMESTATEMENT_85,INCOMESTATEMENT_127,\
                             INCOMESTATEMENT_14,CASHFLOWSTATEMENT_39,CASHFLOWSTATEMENT_59,CASHFLOWSTATEMENT_77,CASHFLOWSTATEMENT_82,CASHFLOWSTATEMENT_86',
                            f'ReportDate={end_date},type=1,Ispandas=1')
        if not isinstance(df_no_param, pd.DataFrame) and df_no_param.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_no_param.ErrorMsg}')
        else:
            df_list.append(df_no_param)

        # 获取主营收入构成(按产品)
        df_mb_sales_cons = c.css(stock_id_result.Codes, "MBSALESCONS", f'ReportDate={end_date},type=2,Ispandas=1')
        if not isinstance(df_mb_sales_cons, pd.DataFrame) and df_mb_sales_cons.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_mb_sales_cons.ErrorMsg}')
        else:
            df_list.append(df_mb_sales_cons.rename(columns={'MBSALESCONS': 'MBSALESCONS_P'}))

        # 获取归属于上市公司股东的扣除非经常性损益后的净利润(调整前)
        df_with_param_1 = c.css(stock_id_result.Codes, "DEDUCTEDINCOME", f'ReportDate={end_date},DataAdjustType=1,Ispandas=1')
        if not isinstance(df_with_param_1, pd.DataFrame) and df_with_param_1.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_with_param_1.ErrorMsg}')
        else:
            df_list.append(df_with_param_1.rename(columns={'DEDUCTEDINCOME': 'DEDUCTEDINCOME_BA'}))

        # 获取归属于上市公司股东的扣除非经常性损益后的净利润(调整后)
        df_with_param_2 = c.css(stock_id_result.Codes, "DEDUCTEDINCOME", f'ReportDate={end_date},DataAdjustType=2,Ispandas=1')
        if not isinstance(df_with_param_2, pd.DataFrame) and df_with_param_2.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_with_param_2.ErrorMsg}')
        else:
            df_list.append(df_with_param_2.rename(columns={'DEDUCTEDINCOME': 'DEDUCTEDINCOME_AA'}))

        # 获取TTM指标
        df_ttm = c.css(stock_id_result.Codes, 'GRTTMR,GCTTMR,ORTTMR,OCTTMR,EXPENSETTMR,GROSSMARGINTTMR,OPERATEEXPENSETTMR,ADMINEXPENSETTMR,FINAEXPENSETTMR,IMPAIRMENTTTMR,\
                       OPERATEINCOMETTMR,INVESTINCOMETTMR,OPTTMR,NONOPERATEPROFITTTMR,EBITTTMR,EBTTTMR,TAXTTMR,PNITTMR,KCFJCXSYJLRTTMR,\
                       NPTTMRP,FVVPALRP,IRTTMRP,IITTMFJVAJVRP,BTAATTMRP,SALESCASHINTTMR,CFOTTMR,CFITTMR,CFFTTMR,CFTTMR,CAPEXR',
                       f'ReportDate={end_date},Ispandas=1')
        if not isinstance(df_ttm, pd.DataFrame) and df_ttm.ErrorCode != 0:
            print(f'failed to get stock financial factors: {df_ttm.ErrorMsg}')
        else:
            df_list.append(df_ttm)

        if len(df_list) > 0:
            # 将每部分按列粘到一起
            df = pd.concat(df_list, axis=1)
            # 日期列修改为end_date
            df.loc[:, 'DATES'] = end_date
            df = df.reset_index().rename(columns=lambda x: EmStockFinFac.__getattribute__(EmStockFinFac, x).name)
            # 更新到db
            self._upload_raw(df, EmStockFinFac.__table__.name)

    def em_stock_daily_info(self, start_date: str, end_date: str, predict_year: int = 0):
        try:
            # 获取区间内所有交易日
            tradedates = c.tradedates(start_date, end_date, "period=1,order=1,market=CNSESH")
            if tradedates.ErrorCode != 0:
                print(f'[em_stock_daily_info] failed to get trade dates: {tradedates.ErrorMsg}')
                return
            if len(tradedates.Data) == 0:
                print(f'[em_stock_daily_info] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return

            # 获取所有A股股票的ID, 功能函数-板块成分
            # http://quantapi.eastmoney.com/Cmd/Sector?from=web
            # 全部A股股票
            stock_id_result = c.sector('001004', end_date)
            if stock_id_result.ErrorCode != 0:
                print(f'failed to get stock id list: {stock_id_result.ErrorMsg}')
                return

            if predict_year == 0:
                predict_year = datetime.datetime.strptime(end_date, '%Y%m%d').date().year

            df_list = []
            for date in tradedates.Data:
                temp_df = c.css(stock_id_result.Codes,
                                'TOTALSHARE,HOLDFROZENAMTACCUMRATIO,ASSETMRQ,EQUITYMRQ,PETTMDEDUCTED,PBLYRN,PSTTM,AHOLDER,ESTPEG',
                                f'EndDate={date},TradeDate={date},PREDICTYEAR={predict_year},Ispandas=1')
                if not isinstance(temp_df, pd.DataFrame) and temp_df.ErrorCode != 0:
                    print(f'failed to get stock daily info: {temp_df.ErrorMsg} (date){date}')
                    return
                # 日期列修改为end_date
                temp_df.loc[:, 'DATES'] = date
                df_list.append(temp_df)

            df = pd.concat(df_list).reset_index().rename(columns=lambda x: EmStockDailyInfo.__getattribute__(EmStockDailyInfo, x).name)
            # 更新到db
            self._upload_raw(df, EmStockDailyInfo.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def em_stock_refinancing(self, start_date: str, end_date: str):
        try:
            data = c.ctr("AdditionPlanInfo",
                         "SECURITYCODE,APPROVENOTICEDATE,PLANNOTICEDDATE,NEWPROGRESS,SUMFINA_T,ATTACHNAME,ADDPURPOSE",
                        f"DateType=2,StartDate={start_date},EndDate={end_date},Process=0")
            if data.ErrorCode != 0:
                print(f'failed to get stock refinancing info: {data.ErrorMsg}')
                return False

            infos: List[List] = []
            for v in data.Data.values():
                # if v[3] in ('董事会批准', '董事会修改', '董事会终止', '股东大会批准', '股东大会修改'):
                #    continue
                infos.append(v)
            df = pd.DataFrame(infos, columns=['SECURITYCODE', 'APPROVENOTICEDATE', 'PLANNOTICEDDATE',
                                              'NEWPROGRESS', 'SUMFINA_T', 'ATTACHNAME', 'ADDPURPOSE'])
            self._upload_raw(df.rename(columns=lambda x: EmStockRefinancing.__getattribute__(EmStockRefinancing, x).name), EmStockRefinancing.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def em_index_val(self, start_date: str, end_date: str):
        try:
            # 获取区间内所有交易日
            tradedates = c.tradedates(start_date, end_date, "period=1,order=1,market=CNSESH")
            if tradedates.ErrorCode != 0:
                print(f'[em_index_val] failed to get trade dates: {tradedates.ErrorMsg}')
                return
            if len(tradedates.Data) == 0:
                print(f'[em_index_val] no trade dates, return immediately (start_date){start_date} (end_date){end_date}')
                return

            # 获取basic table中index_info信息
            index_info_df = BasicDataApi().get_index_info()
            # 下边几个指数特殊处理
            special_list = "399006.SZ,000300.SH,000905.SH,SPX.GI"
            # 获取index_info中满足条件的上述几个指数之外的所有指数
            other_indexes = index_info_df[~index_info_df.loc[:, 'em_id'].isin(special_list.split(',')) & index_info_df.loc[:, 'tag_method'].isin(['PE百分位', 'PB百分位'])]['em_id'].dropna()

            '''
            def transform_tradedates(x: str):
                day, month, year = x.split('/')
                return datetime.date(int(day), int(month), int(year)).strftime('%Y/%m/%d')
            transformed_tradedates = list(map(transform_tradedates, tradedates.Data))
            '''
            df_list = []
            for date in tradedates.Data:
                # 一天一天地获取
                indexes_df = c.csd(special_list,
                                "PETTM,PBMRQ,DIVIDENDYIELD,PSTTM,PCFTTM,ROE,EPSTTM",
                                date, date,
                                "DelType=1,period=1,adjustflag=1,curtype=1,order=1,market=0,Ispandas=1")
                if not isinstance(indexes_df, pd.DataFrame) and indexes_df.ErrorCode != 0:
                    print(f'failed to get index val info: {indexes_df.ErrorMsg} (date){date}')
                    return
                # df_list.append(indexes_df.loc[indexes_df['DATES'].isin(transformed_tradedates), :])
                df_list.append(indexes_df)

                # 两组指数可能获取的日期范围有一些区别，所以分开获取
                other_indexes_df = c.csd(other_indexes.to_list(),
                                        "PETTM,PBMRQ,DIVIDENDYIELD,PSTTM,PCFTTM,ROE,EPSTTM",
                                        date, date,
                                        "DelType=1,period=1,adjustflag=1,curtype=1,order=1,market=0,Ispandas=1")
                if not isinstance(other_indexes_df, pd.DataFrame) and other_indexes_df.ErrorCode != 0:
                    print(f'failed to get index val info: {other_indexes_df.ErrorMsg} (date){date}')
                    return
                # df_list.append(other_indexes_df.loc[other_indexes_df['DATES'].isin(transformed_tradedates), :])
                df_list.append(other_indexes_df)

            # 更新到db
            df = pd.concat(df_list).reset_index().rename(columns=lambda x: EmIndexVal.__getattribute__(EmIndexVal, x).name)
            self._upload_raw(df, EmIndexVal.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False

    def em_tradedates(self, start_date, end_date):
        try:
            tradedates_result = c.tradedates(start_date, end_date,"period=1,order=1,market=CNSESH")
            if tradedates_result.ErrorCode != 0:
                print(f'Failed to get tradedates: {tradedates_result.ErrorMsg}')
                return False

            tradedates = tradedates_result.Data
            df = pd.DataFrame(tradedates, columns=['TRADEDATES'])
            self._upload_raw(df, EmTradeDates.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_stack()
            return False


    def download_all(self, start_date, end_date):
        failed_tasks = []
        # If em_tradedates fails, there is no trading day between start_date and end_date
        # Stop and return
        if not self.em_tradedates(start_date, end_date):
            failed_tasks.append('em_tradedates')
            return failed_tasks

        if not self.em_fund_scale(start_date, end_date):
            failed_tasks.append('em_fund_scale')

        if not self.em_fund_nav(start_date, end_date):
            failed_tasks.append('em_fund_nav')

        if not self.em_index_price(['H00140.SH','H11025.CSI','AU0.SHF'], start_date, end_date):
            failed_tasks.append('em_index_price')

        if not self.em_stock_price(start_date, end_date):
            failed_tasks.append('em_stock_price')
        
        if not self.em_stock_post_price(start_date, end_date):
            failed_tasks.append('em_stock_post_price')

        if not self.em_index_val(start_date, end_date):
            failed_tasks.append('em_index_val')
        
        if not self.em_stock_daily_info(start_date, end_date):
            failed_tasks.append('em_stock_daily_info')

        return failed_tasks


if __name__ == '__main__':
    em = EmRawDataDownloader()
    # em.em_index_val('20200428', '20200428')
    # em.em_fund_nav('20200415', '20200427')
    # em.em_index_price_history('H00140.SH', '20050101', '20200419')
    # em.em_fund_scale_history()
    # em.em_stock_fin_fac('20161231')
    # em.em_stock_info('20200424')
    # em.em_stock_price('20200428', '20200428')
    # em.em_stock_daily_info('20200228', '20200228')
    # em.em_stock_price_history('./surfing/data/fund/raw/stock_price')
    # em.em_stock_refinancing('20170101', '20181231')
    # em.add_columns_to_fin_fac('20190930')
    em.em_fund_scale('20200513', '20200513')




import pandas as pd
import numpy as np
import datetime
import traceback
from ...manager.manager_fund import FundDataManager
from ...manager.score import FundScoreManager
from ...wrapper.mysql import DerivedDatabaseConnector
from ...view.derived_models import FundIndicator

class FundIndicatorProcessor(object):

    TRADING_DAYS_PER_YEAR = 242
    SHORT_TIME_SPAN = TRADING_DAYS_PER_YEAR
    LONG_TIME_SPAN = TRADING_DAYS_PER_YEAR * 3
    LONG_TERM_INDEX = ['hs300', 'csi500', 'mmf']
    MIN_TIME_SPAN = int(TRADING_DAYS_PER_YEAR / 4)#为了延长基金回测范围，评分最低年限3个月

    def __init__(self):
        self.updated_count = {}

    def init(self, start_date, end_date):    
        self._dm = FundDataManager(start_time=start_date, end_time=end_date, score_manager=FundScoreManager())
        self._dm.init(score_pre_calc=False)
        self.start_date = self._dm.start_date
        self.end_date = self._dm.end_date
        self.fund_nav = self._dm.dts.fund_nav.copy()
        self.fund_info = self._dm.dts.fund_info.copy()
        self.index_price = self._dm.dts.index_price.copy()
        
        # 指数日线价格和基金净值 时间轴对齐 ，避免join后因为日期不存在出现价格空值
        # 指数数据历史上出现非交易日有数据，join后，基金净值出现空值，长周期时间序列算指标出空值 
        # 避免对日收益fillna(0)， 造成track error 计算不符合实际
        # 计算逻辑空值填充情况： 1. 只对基金费率空值fillna(0) 
        #                    2. manager fund 里对日线价格ffill
        #                    3. 指数数据按照基金数据重做index后ffill(有基金数据对日子,没有指数数据，比如季报出在季度末，非交易日有基金净值)
        #                    4. 有对超过基金终止日净值赋空值
        end_date = min(self.fund_nav.index[-1], self.index_price.index[-1])
        self.fund_nav = self.fund_nav.loc[:end_date]
        self.index_price = self.index_price[:end_date]
        fund_nav_index = self.fund_nav.index
        self.index_price = self.index_price.reindex(fund_nav_index).fillna(method = 'ffill')
        l1 = self.fund_nav.index.to_list().sort()
        l2 = self.index_price.index.to_list().sort()
        assert l1 == l2, f'date index of index price and fund nav are not identical, l1 {l1} lens {len(l1)}, l2 {l2} lens {len(l2)}'

        fund_to_enddate_dict = self.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        self.fund_to_index_dict = self.fund_info[['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']
        # 超过基金终止日的基金净值赋空
        for fund_id in self.fund_nav.columns:
            fund_end_date = fund_to_enddate_dict[fund_id]
            if self.end_date > fund_end_date:
                self.fund_nav.loc[fund_end_date:,fund_id] = np.nan

        self.index_ret = np.log(self.index_price / self.index_price.shift(1))
        self.fund_ret = np.log(self.fund_nav / self.fund_nav.shift(1))
        self.fund_ret = self.fund_ret.stack().reset_index(level=[0,1]).rename(columns={0:'ret'})        
        self.fund_ret['index_id'] = self.fund_ret.fund_id.apply(lambda x: self.fund_to_index_dict[x])
        self.fund_ret = self.fund_ret.pivot_table(index = ['index_id','datetime'],columns='fund_id',values='ret')
        self.index_ret = np.log(self.index_price / self.index_price.shift(1))
        self.fund_to_index_dict = {fund_id:index_id for fund_id,index_id in self.fund_to_index_dict.items() if fund_id in self.fund_ret.columns}
        self.index_list = self.fund_ret.index.levels[0]
        self.index_fund = { index_id : [fund_idx for fund_idx, index_idx in self.fund_to_index_dict.items() if index_idx == index_id] for index_id in self.index_list}
        
    def get_time_range(self, index_id):
        if index_id in self.LONG_TERM_INDEX:
            return self.LONG_TIME_SPAN
        else:
            return self.SHORT_TIME_SPAN

    ''' 
    尝试过几个版本的回归函数， 计算结果一致
    1. 最早使用的是pyfinance.PandasRollingOLS  可以对一只基金直接返回beta, alpha, 可以设置滚动周期, 运算比较快， 缺点是无法设置min_periods，最小滚动周期
    2. pyfinance.PandasRollingOLS 本质是使用statsmodels.regression.rolling.RollingOLS, 需要手动增加回归常数项，函数下存在最小窗口设置，但是使用无效，
       而且window实际的含义和pandas.rolling.window有区别，导致计算有效值和nan的边界时间不一致
    3. np.polyfit 一元回归，自带回归常数项，可以设置回归最高幂

    为了实现增加最小窗口设置功能的滑动回归， 用pandas.rolling对df.index做滑动计算（验证过不能用datetime index， 只能用整数序列），通过index选取多个列，进行回测
    TODO pandas.rolling 只能返回一列，尝试过一些方法，未果，实质上beta和alpha是同一次回归一起返回的，当前只能分两次计算，增加运算时间
    '''

    def _roll_beta(self, x):
        b_d = int(x[0])
        e_d = int(x[-1])
        df_i = self.df.loc[b_d:e_d,].copy()
        if sum(df_i.fund) == 0:
            return np.Inf
        else:
            res = np.polyfit(y=df_i.fund, x=df_i.benchmark,deg=1)
            return res[0]

    def _roll_alpha(self, x):
        b_d = int(x[0])
        e_d = int(x[-1])
        df_i = self.df.loc[b_d:e_d,].copy()
        if sum(df_i.fund) == 0:
            return np.Inf
        else:
            res = np.polyfit(y=df_i.fund, x=df_i.benchmark,deg=1)
            return res[1] * self.TRADING_DAYS_PER_YEAR

    def upload_derived(self, df, table_name):
        print(table_name)
        df.to_sql(table_name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
        if table_name not in self.updated_count:
            self.updated_count[table_name] = 0
        self.updated_count[table_name] += df.shape[0]

    def calculate(self, index_id_list:str=None):
        self.fund_fee = self.fund_info[['fund_id','manage_fee','trustee_fee']].set_index('fund_id').fillna(0).sum(axis = 1)
        res = []
        index_list = self.index_list if index_id_list is None else index_id_list
        for index_id in index_list:
            time_range = self.get_time_range(index_id)
            fund_list = self.index_fund[index_id]
            fund_ret = self.fund_ret.loc[index_id][fund_list].copy()
            index_ret = self.index_ret[[index_id]]
            for fund_id in fund_list:
                self.df = fund_ret[[fund_id]].join(index_ret).dropna()
                self.df = self.df.rename(columns={index_id:'benchmark',fund_id:'fund'}).reset_index()
                self.df['beta'] = pd.Series(self.df.index).rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).apply(self._roll_beta,raw=True)
                self.df['alpha'] = pd.Series(self.df.index).rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).apply(self._roll_alpha,raw=True)
                self.df['track_err'] = (self.df.fund - self.df.benchmark).rolling(window=time_range, min_periods=self.MIN_TIME_SPAN).std(ddof=1) 
                self.df['fund_id'] = fund_id
                self.df['timespan'] = int(time_range / self.TRADING_DAYS_PER_YEAR)
                self.df['fee_rate'] = self.fund_fee[fund_id]
                self.df = self.df.drop(['fund','benchmark'], axis=1)
                res.append(self.df)
            print(f'index {index_id} fund number : {len(fund_list)} finish' )
        self.result = pd.concat(res, axis=0).replace(np.Inf,None).dropna()

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            start_date_dt = datetime.datetime.strptime(start_date, '%Y%m%d').date()
            end_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()

            start_date = start_date_dt - datetime.timedelta(days = 1150) #3年历史保险起见，多取几天 3*365=1095 
            start_date = datetime.datetime.strftime(start_date, '%Y%m%d')

            self.init(start_date, end_date)
            self.calculate()
            df = self.result[(self.result['datetime'] >= start_date_dt) & (self.result['datetime'] <= end_date_dt)]

            self.upload_derived(df, FundIndicator.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_stack()
            failed_tasks.append('fund_indicator')

        return failed_tasks



from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue
from ...data.struct import AssetTrade, FundTrade, FundPosition, FundWeight
from ...data.struct import AssetTradeParam, FundTradeParam
from . import Helper
import numpy as np
class AssetTrader(Helper):

    def __init__(self, asset_param: AssetTradeParam=None):
        self.asset_param = asset_param or AssetTradeParam()

    def calc_asset_trade(self, dt, 
                               cur_position: AssetPosition, 
                               cur_price: AssetPrice, 
                               target_allocation: AssetWeight):
        cur_mv = AssetValue(prices=cur_price, positions=cur_position)
        tot_mv = cur_mv.sum()

        trades = []
        launch_trade = False
        for index_id, target_weight in target_allocation.__dict__.items():
            if index_id != 'cash':
                target_amt = tot_mv * target_weight
                p = cur_price.__dict__[index_id]
                cur_amt = cur_mv.__dict__[index_id]
                if abs(target_amt - cur_amt) > tot_mv * self.asset_param.MinCountAmtDiff:
                    amount = abs(target_amt - cur_amt)
                    is_buy = target_amt > cur_amt
                    if is_buy:
                        trades.append(AssetTrade(
                            index_id=index_id, 
                            mark_price=p, 
                            amount=amount, 
                            is_buy=is_buy,
                            submit_date=dt
                        ))
                    else:
                        trades.append(AssetTrade(
                            index_id=index_id, 
                            mark_price=p, 
                            volume=amount/p, 
                            is_buy=is_buy,
                            submit_date=dt
                        ))
                launch_trade = launch_trade or abs(target_amt - cur_amt) > tot_mv * self.asset_param.MinActionAmtDiff

        if not launch_trade:
            return cur_position, None
        else:
            trades.sort(key=lambda x: x.is_buy)       
            new_position = cur_position.copy()
            for trd in trades:
                new_position.update(trd)
            return new_position, trades

    def finalize_trade(self, dt, trades: list, 
                            t1_price: AssetPrice, 
                            bt_position: AssetPosition):
        pendings = []
        traded_list = []
        if trades is None or len(trades) == 0:
            return pendings, traded_list
        # TODO: if some trades needs more time
        for trd in trades:
            # TODO: commision calculate
            #trd.commission = ?
            # update position
            trd.trade_price = t1_price.__dict__[trd.index_id]
            trd.volume = trd.volume if trd.volume else (trd.amount / trd.trade_price)
            trd.trade_date = dt
            if not trd.is_buy:
                if not(bt_position.__dict__[trd.index_id] - trd.volume > -1e-8):
                    print(f'trade volume exceeds, adjusted to pos (index_id){trd.index_id} (vol){trd.volume} (is_buy){trd.is_buy} (pos){bt_position.__dict__[trd.index_id]}')
                    trd.volume = bt_position.__dict__[trd.index_id]
                    trd.amount = trd.volume * trd.trade_price
            if self.asset_param.EnableCommission:
                amount = trd.amount if trd.amount else trd.volume * trd.trade_price
                trd.commission = amount * (self.asset_param.PurchaseDiscount * self.asset_param.AssetPurchaseRate[trd.index_id] if trd.is_buy 
                                                else self.asset_param.RedeemDiscount * self.asset_param.AssetRedeemRate[trd.index_id])
            bt_position.update(trd)
            traded_list.append(trd)
        return pendings, traded_list

class FundTrader(AssetTrader):

    def __init__(self, asset_param: AssetTradeParam=None, fund_param: FundTradeParam=None):
        AssetTrader.__init__(self, asset_param=asset_param)
        self.fund_param = fund_param or FundTradeParam()

    def calc_fund_trade(self, dt, fund_weight: FundWeight, cur_fund_position: FundPosition, 
                            cur_fund_nav: dict, 
                            fund_purchase_fees: dict,
                            fund_redeem_fees: dict) -> list:
        new_fund_position = cur_fund_position.copy()
        fund_trades = []
        # return trade list
        fund_tot_mv, cur_fund_wgts = cur_fund_position.calc_mv_n_w(fund_navs=cur_fund_nav)
        all_funds = {}
        # prepare fund candidates and its index_id
        for fund_id, fund_wgt_item in fund_weight.funds.items():
            all_funds[fund_id] = fund_wgt_item.index_id
        for fund_id, fund_pos_item in cur_fund_position.funds.items():
            all_funds[fund_id] = fund_pos_item.index_id
        # calc trade
        for fund_id, index_id in all_funds.items():
            target_fund_amt = fund_weight.get_wgt(fund_id) * fund_tot_mv
            cur_fund_volume = cur_fund_position.get_volume(fund_id) or 0
            p = cur_fund_nav[fund_id]
            cur_fund_amt = cur_fund_volume * p

            if abs(target_fund_amt - cur_fund_amt) > fund_tot_mv * self.fund_param.MinCountAmtDiff or (target_fund_amt == 0 and cur_fund_amt > 0):
                # TODO: commision and 如果是清某一只基金的逻辑，清空可以执行
                is_buy = target_fund_amt > cur_fund_amt
                if is_buy:
                    _trade = FundTrade(
                        fund_id=fund_id,
                        index_id=index_id, 
                        mark_price=p, 
                        amount=abs(target_fund_amt - cur_fund_amt),
                        is_buy=is_buy,
                        submit_date=dt
                    )
                else:
                    _trade = FundTrade(
                        fund_id=fund_id,
                        index_id=index_id, 
                        mark_price=p, 
                        volume=abs(target_fund_amt - cur_fund_amt)/p,
                        is_buy=is_buy,
                        submit_date=dt
                    )
                fund_trades.append(_trade)
                #print(f'(fund){fund_id} (p){p} (amt0){cur_fund_amt} (amt1){target_fund_amt} (idx){index_id} (amt){abs(target_fund_amt - cur_fund_amt)} (direc) {target_fund_amt > cur_fund_amt} ')

        fund_trades.sort(key=lambda x: x.is_buy)
        for _trade in fund_trades:
            new_fund_position.update(_trade)
        return new_fund_position, fund_trades

    def finalize_trade(self, dt, trades: list, 
                            t1_price: AssetPrice, 
                            bt_position: AssetPosition,
                            cur_fund_position: FundPosition, 
                            cur_fund_nav: dict,
                            cur_fund_unit_nav: dict,
                            fund_purchase_fees: dict,
                            fund_redeem_fees: dict):

        if trades is None or len(trades) == 0:
            return [], []
        
        pendings = []
        traded_list = []
        # TODO: if some trades needs more time
        for trd in trades:
            trd.trade_price = cur_fund_nav[trd.fund_id]
            trd.fund_unit_nav = cur_fund_unit_nav[trd.fund_id]
            trd.volume = trd.volume if trd.volume else (trd.amount / trd.trade_price)
            trd.amount = trd.amount if trd.amount else (trd.volume * trd.trade_price)
            trd.fund_unit_volume = trd.amount / trd.fund_unit_nav
            trd.trade_date = dt
            if not trd.is_buy:
                cur_vol = cur_fund_position.get_volume(trd.fund_id)
                if not((cur_vol or 0) - trd.volume > -1e-8):
                    print(f'trade volume exceeds, adjusted to pos (fund_id){trd.fund_id} (vol){trd.volume} (is_buy){trd.is_buy} (pos){cur_vol}')
                    assert cur_vol is not None, 'sell fund with no current position!'
                    trd.volume = cur_vol
                    trd.amount = trd.volume * trd.trade_price
            if self.fund_param.EnableCommission:
                trd.commission = trd.amount * (self.fund_param.PurchaseDiscount * fund_purchase_fees[trd.fund_id] if trd.is_buy 
                                            else self.fund_param.RedeemDiscount * fund_redeem_fees[trd.fund_id])
            trade_status = cur_fund_position.update(trd)
            
            assert trade_status, f"trade failed {trd}"
            if trade_status:
                traded_list.append(trd)
        # get cur asset weight from fund weight
        cur_mv, cur_fund_weight = cur_fund_position.calc_mv_n_w(cur_fund_nav)
        fund_index_dic = {}
        for fund_id, fund_pos_item in cur_fund_position.funds.items():
            fund_index_dic[fund_id] = fund_pos_item.index_id
        asset_wgt = { _ : 0 for _ in set(fund_index_dic.values())}
        for fund_id, wgt_i in cur_fund_weight.items():
            index_id = fund_index_dic[fund_id]
            asset_wgt[index_id] += wgt_i
        # set index position:
        for index_id in asset_wgt:
            asset_p = t1_price.__dict__[index_id]
            amount = cur_mv * asset_wgt[index_id]    
            bt_position.__dict__[index_id] = amount / asset_p
        # fund cash -> asset cash
        bt_position.cash = cur_fund_position.cash
        return pendings, traded_list



import math
import pandas as pd
import pickle as pkl
from ..wrapper.mysql import BasicDatabaseConnector
from ..view.basic_models import *

class BasicDataApi(object):
    def get_trading_day_list(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        TradingDayList
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))

    def get_fund_info(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundInfo
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))


    def get_index_info(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        IndexInfo
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))            

    def get_stock_info(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        StockInfo
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))            

    def get_fund_nav(self, fund_list):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundNav
                ).filter(
                        # We could query all fund_ids at one time
                        FundNav.fund_id.in_(fund_list),
                    )
                    
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))            

    def get_fund_nav_with_date(self, start_date, end_date, fund_list):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundNav.fund_id,
                        FundNav.adjusted_net_value, 
                        FundNav.datetime
                    ).filter(
                        FundNav.fund_id.in_(fund_list),
                        FundNav.datetime >= start_date,
                        FundNav.datetime <= end_date,
                    )
                    
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))        
                
    def get_stock_price(self, stock_list):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        StockPrice
                ).filter(
                        # We could query all fund_ids at one time
                        StockPrice.stock_id.in_(stock_list),
                    )
                    
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))            


    def get_fund_ret(self, fund_list):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundRet
                ).filter(
                        # We could query all fund_ids at one time
                        FundRet.fund_id.in_(fund_list),
                    )
                    
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))            


    def get_index_price(self, index_list):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        IndexPrice
                ).filter(
                        # We could query all fund_ids at one time
                        IndexPrice.index_id.in_(index_list),
                    )
                    
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))            

    def delete_index_price(self, index_id_list, start_date, end_date):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    IndexPrice
                ).filter(
                    IndexPrice.index_id.in_(index_id_list),
                    IndexPrice.datetime >= start_date,
                    IndexPrice.datetime <= end_date,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))
                return None

    def get_fund_nav_adjusted_nv(self, fund_list, start_date, end_date):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundNav.adjusted_net_value,
                        FundNav.fund_id,
                        FundNav.datetime,
                ).filter(
                        FundNav.fund_id.in_(fund_list),
                        FundNav.datetime >= start_date,
                        FundNav.datetime <= end_date,
                    )
                    
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))   


    def get_index_benchmark_data(self, index_list, start_date, end_date):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        IndexPrice.ret,
                        IndexPrice.index_id,
                        IndexPrice.datetime,
                        IndexPrice.close,
                ).filter(
                        IndexPrice.index_id.in_(index_list),
                        IndexPrice.datetime >= start_date,
                        IndexPrice.datetime <= end_date,
                    )
                    
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))   


    def get_fund_list(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundInfo.fund_id
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))

    def get_fund_fee(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundInfo.fund_id,
                        FundInfo.manage_fee,
                        FundInfo.trustee_fee,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))

    def get_fund_asset(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundInfo.fund_id,
                        FundInfo.asset_type,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))

    def get_fund_id_mapping(self):
        fund_id_mapping = {}
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query_results = db_session.query(
                        FundInfo.fund_id,
                        FundInfo.order_book_id,
                        FundInfo.start_date, 
                        FundInfo.end_date
                    ).all()

                for fund_id, order_book_id, start_date, end_date in query_results:
                    if order_book_id not in fund_id_mapping:
                        fund_id_mapping[order_book_id] = []
                    fund_id_mapping[order_book_id].append(
                        {'fund_id': fund_id, 'start_date':start_date, 'end_date': end_date})

            except Exception as e:
                print('Failed to get fund id mapping <err_msg> {}'.format(e))
                return None
                
        return fund_id_mapping

    def get_fund_size(self):
        with BasicDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                        FundSize,
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))



import pandas as pd
import numpy as np
import datetime
import traceback
from ...wrapper.mysql import BasicDatabaseConnector
from ...view.derived_models import FundIndicator
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from .basic_data_part1 import BasicDataPart1
from ...view.basic_models import *

class BasicFundRet(BasicDataPart1):
    # inport end date , none std, 
    W_1 = 5
    M_1 = 20
    M_3 = 60
    M_6 = 121
    Y_1 = 242
    SMALL_NUMBER = 1e-8

    def __init__(self):    
        self.raw_api = RawDataApi()
        self.basic_api = BasicDataApi()
        BasicDataPart1.__init__(self)
        
    def init(self, start_date, end_date):    
        self.start_date = start_date if isinstance(start_date, datetime.date) else datetime.datetime.strptime(start_date,'%Y%m%d').date()
        self.end_date = end_date if isinstance(end_date, datetime.date) else datetime.datetime.strptime(end_date,'%Y%m%d').date()
        self.fund_list = self._fund_info_df.fund_id.tolist()       
        self.fund_nav = self.basic_api.get_fund_nav_with_date(self.start_date, self.end_date, self.fund_list)
        self.fund_scale = self.raw_api.get_em_fund_scale(self.start_date, self.end_date)
        self.index_list = list(set(self._fund_info_df.index_id))
        self._fund_to_enddate_dict = self._fund_info_df[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']  
        self.fund_nav = self.fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').fillna(method='ffill')
        
        # 超过基金终止日的基金净值赋空
        for fund_id in self.fund_nav.columns:
            fund_end_date = self._fund_to_enddate_dict[fund_id]
            if self.end_date > fund_end_date:
                self.fund_nav.loc[fund_end_date:,fund_id] = np.nan
        # 基金规模数据表更新fund_id
        self.fund_scale['fund_id'] = self.fund_scale.apply(
            lambda x: self._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
        self.fund_scale = self.fund_scale.rename(columns={'DATES':'datetime'})[['datetime','FUNDSCALE','fund_id']].pivot_table(index='datetime',values='FUNDSCALE',columns='fund_id')
        self.fund_scale = self.fund_scale.reindex(self.fund_nav.index).fillna(method='ffill').iloc[-1]
        # 基金天数
        _get_days = lambda x : pd.Series(x).count()
        self.days = self.fund_nav.count() - 1
  
    def get_annual_ret(self):
        '''
        累积年化收益
        exp(log(p[-1] / p[0]) / (trade_days / 242)) - 1
        '''
        self.log_total_ret = np.log(self.total_ret)
        ret_yearly = self.days / self.Y_1
        return np.exp(self.log_total_ret / ret_yearly) - 1

    def get_annual_vol(self):
        '''
        累积年化波动率
        (p.shift(1) / p).std(ddof=1) * np.sqrt((days - 1) / year)
        '''
        diff = self.fund_nav.shift(1) / self.fund_nav
        std = diff.std(ddof=1)
        std_yearly = np.sqrt((self.days - 1) / (self.days / self.Y_1))
        return std * std_yearly

    def get_mdd(self):
        '''
        累积最大回撤
        1 - (p / p.cummax()).min()
        '''
        return 1 - (self.fund_nav / self.fund_nav.cummax()).min()

    def calculate_one_date(self):
        self.w1_ret = self.fund_nav.iloc[-1] / self.fund_nav.iloc[-1 - self.W_1 ] - 1
        self.m1_ret = self.fund_nav.iloc[-1] / self.fund_nav.iloc[-1 - self.M_1 ] - 1
        self.m3_ret = self.fund_nav.iloc[-1] / self.fund_nav.iloc[-1 - self.M_3 ] - 1
        self.m6_ret = self.fund_nav.iloc[-1] / self.fund_nav.iloc[-1 - self.M_6 ] - 1
        self.y1_ret = self.fund_nav.iloc[-1] / self.fund_nav.iloc[-1 - self.Y_1 ] - 1
        
        self.total_ret = self.fund_nav.iloc[-1] / self.fund_nav.fillna(method = 'bfill').iloc[0] 
        self.to_date_ret = self.total_ret - 1
        self.annual_ret = self.get_annual_ret()
        self.annual_vol = self.get_annual_vol()
        self.sharpe_ratio = self.annual_ret / self.annual_vol
        self.mdd = self.get_mdd()
        
        data_list = [self.w1_ret, self.m1_ret, self.m3_ret, self.m6_ret, self.y1_ret, self.to_date_ret, self.fund_scale, self.annual_ret, self.sharpe_ratio, self.mdd]
        fac_list = ['w1_ret', 'm1_ret', 'm3_ret', 'm6_ret', 'y1_ret', 'to_date_ret', 'avg_size', 'annual_ret', 'sharpe_ratio', 'mdd']
        self.res = []
        for name, data in zip(fac_list, data_list):
            df = pd.DataFrame(data).rename(columns = {0:name,self.fund_nav.index[-1]:'avg_size'})
            self.res.append(df)
        self.result = pd.concat(self.res, axis=1, sort=True).reset_index().rename(columns={'index':'fund_id'})
        self.result['datetime'] = self.fund_nav.index[-1]
        
    def process_all(self, end_date):
        failed_tasks = []
        try:
            default_begin_date = '20050101'#算累积指标用较久对起始日
            self.init(default_begin_date, end_date)
            self.calculate_one_date()
            self._upload_basic(self.result, FundRet.__table__.name)
            
        except Exception as e:
            print(e)
            traceback.print_stack()
            failed_tasks.append('fund_score')
           
        return failed_tasks